# Copyright (c) OpenMMLab. All rights reserved.
"""Modified from https://github.com/LikeLy-Journey/SegmenTron/blob/master/
segmentron/solver/loss.py (Apache-2.0 License)"""
import torch
import torch.nn as nn


class BinaryDiceLoss(nn.Module):
    """Dice loss of binary class
    Args:
        smooth: A float number to smooth loss, and avoid NaN error, default: 1
        p: Denominator value: \sum{x^p} + \sum{y^p}, default: 2
        predict: A tensor of shape [N, *]
        target: A tensor of shape same with predict
        reduction: Reduction method to apply, return mean over batch if 'mean',
            return sum if 'sum', return a tensor of shape [N,] if 'none'
    Returns:
        Loss tensor according to arg reduction
    Raise:
        Exception if unexpected reduction
    """

    def __init__(self, smooth=1, p=2, reduction="mean"):
        super(BinaryDiceLoss, self).__init__()
        self.smooth = smooth
        self.p = p
        self.reduction = reduction

    def forward(self, predict, target):
        assert predict.shape[0] == target.shape[0], "predict & target batch size don't match"
        predict = predict.contiguous().view(predict.shape[0], -1)
        target = target.contiguous().view(target.shape[0], -1)

        num = torch.sum(torch.mul(predict, target), dim=1) + self.smooth
        den = torch.sum(predict.pow(self.p) + target.pow(self.p), dim=1) + self.smooth

        loss = 1 - num / den

        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        elif self.reduction == "none":
            return loss
        else:
            raise Exception("Unexpected reduction {}".format(self.reduction))


class BalanceCrossEntropyLoss(nn.Module):
    """
    Balanced cross entropy loss.
    Shape:
        - Input: :math:`(N, 1, H, W)`
        - GT: :math:`(N, 1, H, W)`, same shape as the input
        - Mask: :math:`(N, H, W)`, same spatial shape as the input
        - Output: scalar.

    Examples::

        >>> m = nn.Sigmoid()
        >>> loss = nn.BCELoss()
        >>> input = torch.randn(3, requires_grad=True)
        >>> target = torch.empty(3).random_(2)
        >>> output = loss(m(input), target)
        >>> output.backward()
    """

    def __init__(self, negative_ratio=3.0, eps=1e-6):
        super(BalanceCrossEntropyLoss, self).__init__()
        self.negative_ratio = negative_ratio
        self.eps = eps

    def forward(
        self, pred: torch.Tensor, gt: torch.Tensor, mask: torch.Tensor, return_origin=False
    ):
        """
        Args:
            pred: shape :math:`(N, 1, H, W)`, the prediction of network
            gt: shape :math:`(N, 1, H, W)`, the target
            mask: shape :math:`(N, H, W)`, the mask indicates positive regions
        """
        positive = (gt * mask).byte()
        negative = ((1 - gt) * mask).byte()
        positive_count = int(positive.float().sum())
        negative_count = min(int(negative.float().sum()), int(positive_count * self.negative_ratio))
        # loss = nn.functional.binary_cross_entropy(pred, gt, reduction='none')
        loss = nn.functional.binary_cross_entropy_with_logits(pred, gt, reduction="none")
        positive_loss = loss * positive.float()
        negative_loss = loss * negative.float()
        # negative_loss, _ = torch.topk(negative_loss.view(-1).contiguous(), negative_count)
        negative_loss, _ = negative_loss.view(-1).topk(negative_count)

        balance_loss = (positive_loss.sum() + negative_loss.sum()) / (
            positive_count + negative_count + self.eps
        )

        if return_origin:
            return balance_loss, loss
        return balance_loss


class DiceLoss(nn.Module):
    """
    Loss function from https://arxiv.org/abs/1707.03237,
    where iou computation is introduced heatmap manner to measure the
    diversity bwtween tow heatmaps.
    """

    def __init__(self, eps=1e-6):
        super(DiceLoss, self).__init__()
        self.eps = eps

    def forward(self, pred: torch.Tensor, gt, mask, weights=None):
        """
        pred: one or two heatmaps of shape (N, 1, H, W),
            the losses of tow heatmaps are added together.
        gt: (N, 1, H, W)
        mask: (N, H, W)
        """
        return self._compute(pred, gt, mask, weights)

    def _compute(self, pred, gt, mask, weights):
        if pred.dim() == 4:
            pred = pred[:, 0, :, :]
            gt = gt[:, 0, :, :]
        assert pred.shape == gt.shape
        assert pred.shape == mask.shape
        if weights is not None:
            assert weights.shape == mask.shape
            mask = weights * mask
        intersection = (pred * gt * mask).sum()

        union = (pred * mask).sum() + (gt * mask).sum() + self.eps
        loss = 1 - 2.0 * intersection / union
        assert loss <= 1
        return loss


class MaskL1Loss(nn.Module):
    def __init__(self, eps=1e-6):
        super(MaskL1Loss, self).__init__()
        self.eps = eps

    def forward(self, pred: torch.Tensor, gt, mask):
        loss = (torch.abs(pred - gt) * mask).sum() / (mask.sum() + self.eps)
        return loss


class DBLoss(nn.Module):
    def __init__(self, alpha=3.0, beta=1.0, ohem_ratio=3, reduction="mean", eps=1e-6):
        """
        Implement PSE Loss.
        :param alpha: binary_map loss 前面的系数
        :param beta: threshold_map loss 前面的系数
        :param ohem_ratio: OHEM的比例
        :param reduction: 'mean' or 'sum'对 batch里的loss 算均值或求和
        """
        super().__init__()
        assert reduction in ["mean", "sum"], " reduction must in ['mean','sum']"
        self.alpha = alpha
        self.beta = beta
        self.bce_loss = BalanceCrossEntropyLoss(negative_ratio=ohem_ratio)
        self.dice_loss = DiceLoss(eps=eps)
        self.l1_loss = MaskL1Loss(eps=eps)
        self.ohem_ratio = ohem_ratio
        self.reduction = reduction

    def forward(self, pred, batch, use_bce=True):
        shrink_maps = pred[:, 0, :, :]
        threshold_maps = pred[:, 1, :, :]
        binary_maps = pred[:, 2, :, :]

        if use_bce:
            loss_shrink_maps = self.bce_loss(
                pred[:, 3, :, :], batch["shrink_map"], batch["shrink_mask"]
            ) + self.dice_loss(shrink_maps, batch["shrink_map"], batch["shrink_mask"])
        else:
            loss_shrink_maps = self.dice_loss(
                shrink_maps, batch["shrink_map"], batch["shrink_mask"]
            )

        loss_threshold_maps = self.l1_loss(
            threshold_maps, batch["threshold_map"], batch["threshold_mask"]
        )
        metrics = dict(loss_shrink_maps=loss_shrink_maps, loss_threshold_maps=loss_threshold_maps)
        if pred.size()[1] > 2:
            loss_binary_maps = self.dice_loss(
                binary_maps, batch["shrink_map"], batch["shrink_mask"]
            ) + self.bce_loss(binary_maps, batch["shrink_map"], batch["shrink_mask"])
            metrics["loss_binary_maps"] = loss_binary_maps
            loss_all = (
                self.alpha * loss_shrink_maps + self.beta * loss_threshold_maps + loss_binary_maps
            )
            metrics["loss"] = loss_all
        else:
            metrics["loss"] = loss_shrink_maps
        return metrics



"""
This file contains a slightly modified version of the original function,
and serves as the only link to the comic_text_detector project:
https://github.com/dmMaze/comic-text-detector
(License: GNU General Public License v3.0)

Should a newer version of the comic_text_detector project be released,
replacing the files in the pcleaner/comic_text_detector folder with the
newer version should be sufficient to update the code, provided that the
coupling between this function and the rest of the code is not broken.

The code present in the pcleaner/comic_text_detector folder is a
slimmed-down version of the original project, containing only the
necessary files to run the model and the function below.
"""

import json
import random
import string
from pathlib import Path

from tqdm import tqdm

from .comic_text_detector.inference import TextDetector
from .comic_text_detector.utils.io_utils import imread, imwrite, find_all_imgs, NumpyEncoder
from .comic_text_detector.utils.textmask import REFINEMASK_ANNOTATION


def model2annotations(model_path: Path, img_paths: str | Path | list[str | Path], save_dir: Path):
    """
    Run the model on a directory of images and produce the following
    for each image inside the save_dir directory:
    - A copy of the original image as a .png file.
    - A .png file containing the text mask, filename: <image_name>_mask.png.
    - A .json file containing each box of text, as well as other metadata,
      filename: <image_name>.json.

    For this modified version, include the image name and mask name in the
    json file.

    :param model_path: Path to the model file. This ends either in .pt or .onnx (torch or cv2 format).
    :param img_paths: Path or a list of paths to an image or directory of images.
    :param save_dir: Path to the directory where the results will be saved.
    :return:
    """
    if isinstance(img_paths, str | Path):
        img_paths = [img_paths]
    device = "cuda" if model_path.suffix == ".pt" else "cpu"
    print(f"Using device for text detection model: {device}")
    model = TextDetector(model_path=str(model_path), input_size=1024, device=device)
    img_list = []

    for img_dir in img_paths:
        if Path(img_dir).is_dir():
            img_list.extend(find_all_imgs(img_dir, abs_path=True))
        elif Path(img_dir).is_file():
            img_list.append(img_dir)
        else:
            raise FileNotFoundError(f"Image path {img_dir} does not exist.")

    if not img_list:
        print("No images found.")
        return

    # Make a batch prefix, which is a random string of uppercase letters.
    # This is used to avoid overwriting files with the same name, even
    # between different runs of the program.
    batch_prefix = get_random_uppercase_string(4)
    digits = len(str(len(img_list) - 1))

    for index, img_path in enumerate(tqdm(img_list)):
        img = imread(img_path)
        img_path = Path(img_path).absolute()

        # Prepend an index to prevent name clobbering between different files.
        prefix = f"{batch_prefix}-{index:0{digits}}_"

        img_name = prefix + img_path.stem
        maskname = img_name + "_mask.png"

        # Make names absolute paths.
        img_name = str((save_dir / (img_name + ".png")).absolute())
        maskname = str((save_dir / maskname).absolute())

        mask, mask_refined, blk_list = model(
            img, refine_mode=REFINEMASK_ANNOTATION, keep_undetected_mask=True
        )
        blk_xyxy = []
        blk_dict_list = []
        for blk in blk_list:
            blk_xyxy.append(blk.xyxy)
            blk_dict_list.append(blk.to_dict())

        # Inject the img_name.png and mask name and original path into the json.
        data = {
            "image_path": img_name,
            "mask_path": maskname,
            "original_path": str(img_path),
            "blk_list": blk_dict_list,
        }

        with open(Path(img_name).with_suffix(".json"), "w", encoding="utf8") as f:
            json.dump(data, f, ensure_ascii=False, cls=NumpyEncoder, indent=4)
        imwrite(img_name, img)
        imwrite(maskname, mask_refined)


def get_random_uppercase_string(length: int = 4) -> str:
    """
    Return a random string of uppercase letters of the given length.

    :param length: The length of the string to return.
    :return: A random string of uppercase letters of the given length.
    """
    letters = string.ascii_uppercase
    random_indices = random.sample(range(len(letters)), length)
    return "".join(map(lambda i: letters[i], random_indices))



import colorsys
from pathlib import Path

import numpy as np
import cv2
import scipy
from PIL import Image, ImageFilter
from logzero import logger

import pcleaner.structures as st
import pcleaner.config as cfg


class BlankMaskError(Exception):
    pass


def convert_mask_to_rgba(mask: Image, color: int | tuple[int, int, int, int] = 255) -> Image:
    """
    Convert "1" mask to "RGBA" mask, where black has an alpha of 0 and white has an alpha of 255.

    :param mask: The mask to convert.
    :param color: [Optional] The color to make the mask, either 1 int (luminance) or 4 ints (rgba).
    :return: The converted mask.
    """
    array_1 = np.array(mask)  # Mode "1" mask.
    color_tuple = (color, color, color, 255) if isinstance(color, int) else color
    # RGBA mask, filled with 0s.
    array_rgba = np.zeros((*array_1.shape, 4), dtype=np.uint8)  # Mode "RGBA" mask.
    # For anything not black in the original mask, set it to (color, color, color, 255).
    array_rgba[array_1 != 0] = color_tuple
    return Image.fromarray(array_rgba)


def apply_debug_filter_to_mask(
    img: Image, color: tuple[int, int, int, int] = (108, 30, 240, 127)
) -> Image:
    """
    Make the mask purple and 50% transparent.
    The mask has either transparent or opaque pixels.

    :param img: image to apply the filter to. Mode: "1" or "RGBA"
    :param color: color to make the mask.
    :return: the filtered image. Mode: "RGBA"
    """

    if img.mode == "1":
        return convert_mask_to_rgba(img, color)
    elif img.mode == "RGBA":
        array_rgba = np.array(img)
        array_rgba[array_rgba[:, :, 3] != 0] = color
        return Image.fromarray(array_rgba)
    else:
        raise ValueError(f"Unknown mode: {img.mode}")


def mask_intersection(mask1: Image, mask2: Image) -> Image:
    """
    Cut out the parts of mask1 that are not in mask2.

    :param mask1: The base mask.
    :param mask2: The mask to cut out of the base mask.
    :return: The intersection mask.
    """
    return Image.composite(mask1, Image.new("1", mask1.size, 0), mask2)


def compose_masks(
    base_size: tuple[int, int], masks_with_pos: list[tuple[Image, tuple[int, int]]]
) -> Image:
    """
    Compose a list of masks into one mask.

    :param base_size: The size of the base image.
    :param masks_with_pos: A list of masks and their positions.
    :return: The composed mask.
    """
    # Create a blank mask.
    base_mask = Image.new("1", base_size, 0)

    # Paste each mask onto the base mask.
    for mask, pos in masks_with_pos:
        base_mask.paste(mask, pos, mask)

    return base_mask


def visualize_mask_fitments(
    base_image: Image, mask_fitments: list[st.MaskFittingResults], output_path: Path
) -> None:
    """
    Stack each mask on top of the base image and save the result.
    To make each mask visible, interpolate across a range of colors,
    adding each mask with some opacity.

    :param base_image: The base image to stack the masks on top of.
    :param mask_fitments: The data of the masks to apply.
    :param output_path: The path to save the image to.
    """
    # Compose all masks of the same index into one mask.
    # Each fitment result contains a list of the sub-masks and their positions.
    masks = [
        compose_masks(
            base_image.size,
            [
                (mask_fitment.debug_masks[index], mask_fitment.mask_coords)
                for mask_fitment in mask_fitments
            ],
        )
        for index in range(len(mask_fitments[0].debug_masks))
    ]

    # Don't modify the original image.
    base_image = base_image.copy()

    # Generate colors for each mask.
    hues = map(lambda i: (0.15 * i) % 1, range(len(masks)))

    # Convert the hues to RGB tuples and scale the values to the range [0, 255]
    colors = (
        (int(r * 255), int(g * 255), int(b * 255), 255)
        for r, g, b in (colorsys.hsv_to_rgb(hue, 1.0, 1.0) for hue in hues)
    )

    # Make each a different color, then overlay them all on top of each other.
    # Flip the order of the masks, so the largest mask comes first and the rest get pasted on top of it.
    colored_masks = (
        apply_debug_filter_to_mask(mask, color) for mask, color in zip(reversed(masks), colors)
    )

    # Combine the masks.
    combined_mask = Image.new("RGBA", base_image.size)
    for mask in colored_masks:
        combined_mask.paste(mask, (0, 0), mask)

    # Limit the alpha channel to 60% = 153.
    alpha_mask = combined_mask.split()[3]
    alpha_mask = alpha_mask.point(lambda x: min(x, 153))
    base_image.paste(combined_mask, (0, 0), alpha_mask)

    # Save the image.
    base_image.save(output_path)


def make_mask_steps_convolution(mask: Image, growth_step: int = 2, steps: int = 11) -> list[Image]:
    """
    Use convolution to make various sizes of the original mask.
    In each step, the outline of the mask is grown by growth_step pixels.

    :param mask: The mask to make steps of.
    :param growth_step: The amount to grow the mask by each step.
    :param steps: The number of steps to make.
    :return:
    """
    masks = []
    grow_size = growth_step * 2
    mask = mask.convert("L")
    mask_array = np.array(mask, dtype=np.uint8)
    # Make a convolution kernel.
    kernel = np.ones((grow_size + 1, grow_size + 1))
    # Remove the corner pixels from the kernel, this will give 45 degree corners.
    kernel[0, 0] = 0
    kernel[0, -1] = 0
    kernel[-1, 0] = 0
    kernel[-1, -1] = 0
    # Pad the image so that the convolution can handle the edges.
    padded_mask = np.pad(mask_array, ((grow_size, grow_size), (grow_size, grow_size)), mode="edge")

    for _ in range(steps):
        # Apply convolve2d to the image.
        padded_mask = scipy.signal.convolve2d(padded_mask, kernel, mode="same")
        cropped_mask = padded_mask[grow_size:-grow_size, grow_size:-grow_size]
        masks.append(
            Image.fromarray(np.where(cropped_mask > 0, 255, 0).astype(np.uint8)).convert("1")
        )

    # Return them ordered from smallest to largest.
    return masks


def border_std_deviation(base: Image, mask: Image, off_white_threshold: int) -> tuple[float, int]:
    """
    Calculate the border uniformity of a mask.
    For this, find the edge pixels of the mask and then calculate the median color of the pixels around them.
    Also calculate the standard deviation of the colors around the edge pixels.

    In case the mask is blank, return a very high standard deviation.

    :raises BlankMaskError: If one or more masks are blank, meaning that the box is not valid.

    :param base: The base image. Mode: "L"
    :param mask: The mask to calculate the border uniformity of. Mode: "1"
    :param off_white_threshold: The threshold for a pixel to be considered off-white.
    :return: The border uniformity as the standard deviation and the median color of the border.
    """
    # Transform the mask into its edge pixels.
    mask = mask.filter(ImageFilter.FIND_EDGES)
    # Collect all pixels from the base image where the mask has a value of 1.
    # Use numpy for efficiency.
    base_data = np.array(base)
    mask_data = np.array(mask)

    border_pixels = base_data[mask_data == 1]
    # Calculate the number of pixels in the border.
    num_pixels = len(border_pixels)
    if num_pixels == 0:
        # We received an empty mask.
        raise BlankMaskError
    # Calculate the standard deviation of the border pixels.
    std = float(np.std(border_pixels))
    # Get the average color of the border pixels, round to highest integer.
    median_color = int(np.median(border_pixels))
    # logger.debug(f"Border uniformity: {std} ({num_pixels} pixels), median color: {median_color}")
    # Round up any color over the threshold to 255. This should prevent slight off-white colors from sneaking in.
    if median_color > off_white_threshold:
        median_color = 255

    return std, median_color


def cut_out_box(image: Image, box: tuple[int, int, int, int]) -> Image:
    """
    Cut out a box from an image.

    :param image: The image to cut out the box from.
    :param box: The box to cut out.
    :return: The cut out box.
    """
    return image.crop(box)


def cut_out_mask(
    mask: Image,
    box: tuple[int, int, int, int],
    target_shape: tuple[int, int] = None,
    x_padding_offset: int = 0,
    y_padding_offset: int = 0,
) -> Image:
    """
    Cut out a box from a mask.
    If a target shape is given, the mask is padded to that shape with black.
    Use the offset to move the mask to the correct position when padding.

    :param mask: Mask image in mode "1".
    :param box: The box to cut out.
    :param target_shape: [Optional] The target shape to pad the mask to.
    :param x_padding_offset: [Optional] The x offset of the mask when padding.
    :param y_padding_offset: [Optional] The y offset of the mask when padding.
    :return: The cut out mask.
    """
    # Cut out the box.
    mask = mask.crop(box)
    # If a target shape is given, pad the mask to that shape.
    if target_shape is not None:
        # Create a black image with the target shape.
        padded_mask = Image.new("1", target_shape, 0)
        # Paste the mask into the black image.
        padded_mask.paste(mask, (x_padding_offset, y_padding_offset))
        mask = padded_mask

    return mask


def pick_best_mask(
    base: Image,
    precise_mask: Image,
    box_mask: Image,
    masking_box: tuple[int, int, int, int],
    reference_box: tuple[int, int, int, int],
    cleaner_conf: cfg.CleanerConfig,
    analytics_page_path: Path,
) -> None | st.MaskFittingResults:
    """
    Generate various sizes of the precise mask and pick the best one from
    among them, and the box mask.

    Return the best mask, the median color of the border, and the coordinates of the mask.
    If the best standard deviation is too high, return None for the mask and color.

    If the precise mask is blank, return None. This happens when the AI picked up some noise.

    "Best" is defined as having the lowest std deviation at the greatest mask size.
    Here, the standard deviation among the color of pixels along the edge of the mask is calculated.
    If they're all the same color, we can safely assume that the mask is a good fit.

    It also returns analytics, containing the standard deviation of the best mask and its index.
    These can be used to gauge how good the mask generation was.


    :param base: The base image.
    :param precise_mask: The precise mask.
    :param box_mask: The box mask.
    :param masking_box: The box to cut the mask out of.
    :param reference_box: The box to cut the base image out of.
    :param cleaner_conf: The cleaner config.
    :param analytics_page_path: The path to the original image for the analytics.
    :return: The best mask and what color to make it and the box (along with analytics). If no best
        mask was found, return None for the mask and color.
    """
    # Calculate offset coords between the reference box and the mask box.
    # The mask box was grown by n pixels in each direction to make the reference box,
    # but it could not exceed the image dimensions.
    # These relative coords are used to paste the mask onto the base image.
    x_offset = masking_box[0] - reference_box[0]
    y_offset = masking_box[1] - reference_box[1]

    # Replace all given images with their corresponding cutouts.
    base = cut_out_box(base, reference_box)
    precise_mask = cut_out_mask(precise_mask, masking_box, base.size, x_offset, y_offset)

    # Check that the precise mask is not blank.
    if precise_mask.getbbox() is None:
        # This means the box was probably just noise, so abort picking a mask.
        logger.warning("Found an empty mask, this is likely due to name collisions.")
        return None

    box_mask = cut_out_mask(box_mask, masking_box, base.size, x_offset, y_offset)

    # Generate masks of various sizes for the precise mask, then add the box mask to the list.
    # The generated masks are in ascending size order.
    masks = make_mask_steps_convolution(
        precise_mask, cleaner_conf.mask_growth_step_pixels, cleaner_conf.mask_growth_steps
    )
    masks.append(box_mask)

    # Calculate the border uniformity of each mask.
    # Border deviations: (std deviation, median color)
    border_deviations: list[tuple[Image, int]] = []
    for mask in masks:
        try:
            border_deviations.append(
                border_std_deviation(base, mask, cleaner_conf.off_white_max_threshold)
            )
        except BlankMaskError:
            return None
    # Find the best mask.
    best_mask = None
    lowest_border_deviation = None
    lowest_deviation_color = None
    for i in range(len(masks)):
        if i == 0 or border_deviations[i][0] <= (
            lowest_border_deviation * (1 - cleaner_conf.mask_improvement_threshold)
        ):
            lowest_border_deviation = border_deviations[i][0]
            lowest_deviation_color = border_deviations[i][1]
            best_mask = masks[i]

    # If the std deviation is too high, return None.
    if lowest_border_deviation > cleaner_conf.mask_max_standard_deviation:
        return st.MaskFittingResults(
            best_mask=None,
            median_color=lowest_deviation_color,
            mask_coords=reference_box[:2],
            analytics_page_path=analytics_page_path,
            analytics_mask_index=masks.index(best_mask),
            analytics_std_deviation=lowest_border_deviation,
            mask_box=masking_box,
            debug_masks=masks,
        )
    return st.MaskFittingResults(
        best_mask=best_mask,
        median_color=lowest_deviation_color,
        mask_coords=reference_box[:2],
        analytics_page_path=analytics_page_path,
        analytics_mask_index=masks.index(best_mask),
        analytics_std_deviation=lowest_border_deviation,
        mask_box=masking_box,
        debug_masks=masks,
    )


def combine_best_masks(
    image_size: tuple[int, int], mask_fitments: list[st.MaskFittingResults]
) -> Image:
    """
    Merge the masks together into a single mask in RGBA mode to preserve the black/white information.
    Masks are in mode "1".

    :param image_size: The size of the image to make the mask for.
    :param mask_fitments: The data of the masks to apply.
    :return: The combined mask. Image mode: "RGBA"
    """
    combined_mask = Image.new("RGBA", image_size, (0, 0, 0, 0))
    # Convert the masks to RGBA mode.
    # Then paste the mask onto the combined mask.
    for mask_fitment in mask_fitments:
        mask, color, coords = mask_fitment.mask_data
        mask = convert_mask_to_rgba(mask, color)
        combined_mask.paste(mask, coords, mask)

    return combined_mask


def combine_noise_masks(
    image_size: tuple[int, int], masks_with_coords: list[tuple[Image, tuple[int, int]]]
) -> Image:
    """
    Merge the noise masks together into a single mask in RGBA mode.
    The noise mask may be blank if all masks either failed or fitted too well.

    :param image_size: The size of the image to make the mask for.
    :param masks_with_coords: The masks to combine with their coordinates.
    :return: The combined mask. Image mode: "RGBA" and whether the mask is blank.
    """
    combined_noise_mask = Image.new("RGBA", image_size, (0, 0, 0, 0))
    # Then paste the mask onto the combined mask.
    for mask, coords in masks_with_coords:
        combined_noise_mask.paste(mask, coords)

    return combined_noise_mask


def denoise(
    image: Image,
    colored: bool = False,
    filter_strength: int = 10,
    color_filter_strength: int = 10,
    template_window_size: int = 7,
    search_window_size: int = 21,
) -> Image:
    """
    Denoise the image using cv2.

    :param image: The image to denoise.
    :param colored: Whether the image is colored or not, to use the appropriate algorithm.
    :param filter_strength: The strength of the filter. Higher values mean more denoising.
    :param color_filter_strength: The strength of the color filter. Higher values mean more denoising.
    :param template_window_size: Size in pixels of the template patch that is used to compute weights. Should be odd.
    :param search_window_size: Size in pixels of the window that is used to compute weighted average for given pixel. Should be odd.
    :return: The denoised image.
    """
    if colored:
        return Image.fromarray(
            cv2.fastNlMeansDenoisingColored(
                np.array(image),
                None,
                filter_strength,
                color_filter_strength,
                template_window_size,
                search_window_size,
            )
        )
    else:
        return Image.fromarray(
            cv2.fastNlMeansDenoising(
                np.array(image), None, filter_strength, template_window_size, search_window_size
            )
        )


def grow_mask(mask: Image, size: int) -> Image:
    """
    Grow the mask by the given amount of pixels.
    Use convolutions to draw a thicker outline.

    :param mask: The mask to grow. Mode: "1"
    :param size: The amount of pixels to grow the mask by.
    :return: The grown mask. Mode: "1"
    """
    grow_size = size * 2
    kernel_size = grow_size + 1
    mask = mask.convert("L")
    mask_array = np.array(mask, dtype=np.uint8)
    # Make a convolution kernel.
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    # Remove the corner pixels from the kernel, this will give 45 degree corners.
    kernel[0, 0] = 0
    kernel[0, -1] = 0
    kernel[-1, 0] = 0
    kernel[-1, -1] = 0
    # Pad the image so that the convolution can handle the edges.
    padded_mask = np.pad(mask_array, ((grow_size, grow_size), (grow_size, grow_size)), mode="edge")

    # Apply convolve2d to the image.
    padded_mask = scipy.signal.convolve2d(padded_mask, kernel, mode="same")
    cropped_mask = padded_mask[grow_size:-grow_size, grow_size:-grow_size]
    return Image.fromarray(np.where(cropped_mask > 0, 255, 0).astype(np.uint8)).convert("1")


def fade_mask_edges(mask: Image, fade_radius: int) -> Image:
    """
    Fade the edges of the mask by the given amount of pixels.

    :param mask: mask to fade. Mode: "1"
    :param fade_radius: The amount of pixels to fade the mask by.
    :return: The faded mask. Mode: "L"
    """
    # Use gaussian blur to fade the edges, perform it in L mode.
    mask = mask.convert("L").filter(ImageFilter.GaussianBlur(fade_radius))
    return mask


def generate_noise_mask(
    image: Image, mask: Image, box: tuple[int, int, int, int], denoiser_conf: cfg.DenoiserConfig
) -> tuple[Image, tuple[int, int]]:
    """
    Cut out the image and mask using the box, then denoise the image cutout.
    Then grow the mask and fade it's edges.
    Create an alpha composite of the denoised image and the mask.

    :param image: The image to cut out and denoise.
    :param mask: The mask.
    :param box: The box to cut out.
    :param denoiser_conf: The denoiser config.
    :return: The noise mask and it's coordinates.
    """

    # Cut out the image and mask.
    image_cutout = image.crop(box)
    mask_cutout = mask.crop(box)
    mask_cutout = grow_mask(mask_cutout, denoiser_conf.noise_outline_size)
    mask_faded = fade_mask_edges(mask_cutout, denoiser_conf.noise_fade_radius)

    # Denoise the image cutout.
    denoised_image_cutout = denoise(
        image_cutout,
        denoiser_conf.colored_images,
        denoiser_conf.filter_strength,
        denoiser_conf.color_filter_strength,
        denoiser_conf.template_window_size,
        denoiser_conf.search_window_size,
    )

    # Add alpha channels.
    denoised_image_cutout.putalpha(mask_faded)

    return denoised_image_cutout, box[:2]



import numpy as np
import cv2
import random


def hex2bgr(hex):
    gmask = 254 << 8
    rmask = 254
    b = hex >> 16
    g = (hex & gmask) >> 8
    r = hex & rmask
    return np.stack([b, g, r]).transpose()


def union_area(bboxa, bboxb):
    x1 = max(bboxa[0], bboxb[0])
    y1 = max(bboxa[1], bboxb[1])
    x2 = min(bboxa[2], bboxb[2])
    y2 = min(bboxa[3], bboxb[3])
    if y2 < y1 or x2 < x1:
        return -1
    return (y2 - y1) * (x2 - x1)


def get_yololabel_strings(clslist, labellist):
    content = ""
    for cls, xywh in zip(clslist, labellist):
        content += str(int(cls)) + " " + " ".join([str(e) for e in xywh]) + "\n"
    if len(content) != 0:
        content = content[:-1]
    return content


# 4 points bbox to 8 points polygon
def xywh2xyxypoly(xywh, to_int=True):
    xyxypoly = np.tile(xywh[:, [0, 1]], 4)
    xyxypoly[:, [2, 4]] += xywh[:, [2]]
    xyxypoly[:, [5, 7]] += xywh[:, [3]]
    if to_int:
        xyxypoly = xyxypoly.astype(np.int64)
    return xyxypoly


def xyxy2yolo(xyxy, w: int, h: int):
    if xyxy == [] or xyxy == np.array([]) or len(xyxy) == 0:
        return None
    if isinstance(xyxy, list):
        xyxy = np.array(xyxy)
    if len(xyxy.shape) == 1:
        xyxy = np.array([xyxy])
    yolo = np.copy(xyxy).astype(np.float64)
    yolo[:, [0, 2]] = yolo[:, [0, 2]] / w
    yolo[:, [1, 3]] = yolo[:, [1, 3]] / h
    yolo[:, [2, 3]] -= yolo[:, [0, 1]]
    yolo[:, [0, 1]] += yolo[:, [2, 3]] / 2
    return yolo


def yolo_xywh2xyxy(xywh: np.array, w: int, h: int, to_int=True):
    if xywh is None:
        return None
    if len(xywh) == 0:
        return None
    if len(xywh.shape) == 1:
        xywh = np.array([xywh])
    xywh[:, [0, 2]] *= w
    xywh[:, [1, 3]] *= h
    xywh[:, [0, 1]] -= xywh[:, [2, 3]] / 2
    xywh[:, [2, 3]] += xywh[:, [0, 1]]
    if to_int:
        xywh = xywh.astype(np.int64)
    return xywh


def rotate_polygons(center, polygons, rotation, new_center=None, to_int=True):
    if new_center is None:
        new_center = center
    rotation = np.deg2rad(rotation)
    s, c = np.sin(rotation), np.cos(rotation)
    polygons = polygons.astype(np.float32)

    polygons[:, 1::2] -= center[1]
    polygons[:, ::2] -= center[0]
    rotated = np.copy(polygons)
    rotated[:, 1::2] = polygons[:, 1::2] * c - polygons[:, ::2] * s
    rotated[:, ::2] = polygons[:, 1::2] * s + polygons[:, ::2] * c
    rotated[:, 1::2] += new_center[1]
    rotated[:, ::2] += new_center[0]
    if to_int:
        return rotated.astype(np.int64)
    return rotated


def letterbox(
    im, new_shape=(640, 640), color=(0, 0, 0), auto=False, scaleFill=False, scaleup=True, stride=128
):
    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if not isinstance(new_shape, tuple):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    # dw /= 2  # divide padding into 2 sides
    # dh /= 2
    dh, dw = int(dh), int(dw)

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, 0, dh, 0, dw, cv2.BORDER_CONSTANT, value=color)  # add border
    return im, ratio, (dw, dh)


def resize_keepasp(im, new_shape=640, scaleup=True, interpolation=cv2.INTER_LINEAR, stride=None):
    shape = im.shape[:2]  # current shape [height, width]

    if new_shape is not None:
        if not isinstance(new_shape, tuple):
            new_shape = (new_shape, new_shape)
    else:
        new_shape = shape

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))

    if stride is not None:
        h, w = new_unpad
        if new_shape[0] % stride != 0:
            new_h = (stride - (new_shape[0] % stride)) + h
        else:
            new_h = h
        if w % stride != 0:
            new_w = (stride - (w % stride)) + w
        else:
            new_w = w
        new_unpad = (new_h, new_w)

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=interpolation)
    return im


def expand_textwindow(img_size, xyxy, expand_r=8, shrink=False):
    im_h, im_w = img_size[:2]
    x1, y1, x2, y2 = xyxy
    w = x2 - x1
    h = y2 - y1
    paddings = int(round((max(h, w) * 0.25 + min(h, w) * 0.75) / expand_r))
    if shrink:
        paddings *= -1
    x1, y1 = max(0, x1 - paddings), max(0, y1 - paddings)
    x2, y2 = min(im_w - 1, x2 + paddings), min(im_h - 1, y2 + paddings)
    return [x1, y1, x2, y2]


def draw_connected_labels(
    num_labels, labels, stats, centroids, names="draw_connected_labels", skip_background=True
):
    labdraw = np.zeros((labels.shape[0], labels.shape[1], 3), dtype=np.uint8)
    max_ind = 0
    if isinstance(num_labels, int):
        num_labels = range(num_labels)

    # for ind, lab in enumerate((range(num_labels))):
    for lab in num_labels:
        if skip_background and lab == 0:
            continue
        randcolor = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
        labdraw[np.where(labels == lab)] = randcolor
        maxr, minr = 0.5, 0.001
        maxw, maxh = stats[max_ind][2] * maxr, stats[max_ind][3] * maxr
        minarea = labdraw.shape[0] * labdraw.shape[1] * minr

        stat = stats[lab]
        bboxarea = stat[2] * stat[3]
        if stat[2] < maxw and stat[3] < maxh and bboxarea > minarea:
            pix = np.zeros((labels.shape[0], labels.shape[1]), dtype=np.uint8)
            pix[np.where(labels == lab)] = 255

            rect = cv2.minAreaRect(cv2.findNonZero(pix))
            box = np.int0(cv2.boxPoints(rect))
            labdraw = cv2.drawContours(labdraw, [box], 0, randcolor, 2)
            labdraw = cv2.circle(
                labdraw,
                (int(centroids[lab][0]), int(centroids[lab][1])),
                radius=5,
                color=(random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)),
                thickness=-1,
            )

    cv2.imshow(names, labdraw)
    return labdraw



from pathlib import Path

from PIL import Image
from logzero import logger

import pcleaner.image_ops as ops
import pcleaner.structures as st


def clean_page(c_data: st.CleanerData) -> list[tuple[Path, bool, int, float]]:
    """
    Do all the shit and return analytics for mask fitting.

    :param c_data: All the data needed for the cleaning process.
    :return: Analytics, consisting of the json file path, and a list of whether the mask was generated, the index of
        the best mask, and the border uniformity of the best mask for each box.
    """

    page_data = st.PageData.from_json(c_data.json_path.read_text())

    # Make a shorter alias.
    cleaner_conf = c_data.cleaner_config

    original_img_path_as_png = Path(page_data.original_path).with_suffix(
        ".png"
    )  # Make sure all derived file names are .png.
    cache_out_path = c_data.cache_dir / original_img_path_as_png.name
    logger.debug(f"Masking {cache_out_path.name}...")

    def save_mask(img, name_suffix):
        if c_data.show_masks:
            img.save(cache_out_path.with_stem(cache_out_path.stem + name_suffix))

    # Load the base image.
    base_image = Image.open(page_data.image_path)

    # Make a box mask. These will serve as basic masks as opposed to the AI generated precise ones.
    box_mask = page_data.make_box_mask(base_image.size, st.BoxType.EXTENDED_BOX)

    save_mask(box_mask, "_box_mask")

    # Load the precise AI-generated mask.
    mask = Image.open(page_data.mask_path)
    mask = mask.convert("1", dither=Image.NONE)  # Convert to bitmap.

    # Cut the mask to the box mask, leaving only the intersection.
    cut_mask = ops.mask_intersection(mask, box_mask)

    save_mask(cut_mask, "_cut_mask")

    # Use the merged extended boxes to guard against overlapping boxes after having extended them.
    # Generate varying sizes of the precise mask and then rank them based on border uniformity.
    mask_fitments: list[st.MaskFittingResults] = [
        ops.pick_best_mask(
            base=base_image,
            precise_mask=cut_mask,
            box_mask=box_mask,
            masking_box=masking_box,
            reference_box=reference_box,
            cleaner_conf=cleaner_conf,
            analytics_page_path=Path(original_img_path_as_png),
        )
        for masking_box, reference_box in zip(
            page_data.merged_extended_boxes, page_data.reference_boxes
        )
    ]
    # Take out all entries that are None, since these masks were false positives.
    mask_fitments = [m for m in mask_fitments if m is not None]
    # Take out the analytics.
    # The analytics consist of (mask could be found: bool, mask index: int, border deviation: float)
    analytics: list[tuple[Path, bool, int, float]] = [m.analytics for m in mask_fitments]

    # Remove entries if they failed.
    best_masks = [m for m in mask_fitments if not m.failed]

    combined_mask = ops.combine_best_masks(base_image.size, best_masks)

    if c_data.show_masks:
        # Save the masks in the debug folder.
        ops.visualize_mask_fitments(
            base_image, mask_fitments, cache_out_path.with_stem(cache_out_path.stem + "_masks")
        )

        # Output the result with the debug filter.
        combined_mask_debug = ops.apply_debug_filter_to_mask(
            combined_mask, cleaner_conf.debug_mask_color
        )
        base_image_copy = base_image.copy()
        base_image_copy.paste(combined_mask_debug, (0, 0), combined_mask_debug)
        save_mask(base_image_copy, "_with_masks")

    base_image.paste(combined_mask, (0, 0), combined_mask)

    # Save the combined mask. This will be used for denoising.
    combined_mask_path = cache_out_path.with_stem(cache_out_path.stem + "_combined_mask")
    combined_mask.save(combined_mask_path)
    cleaned_image_path = cache_out_path.with_stem(cache_out_path.stem + "_cleaned")
    base_image.save(cleaned_image_path)
    save_denoising_data(
        original_img_path_as_png,
        cleaned_image_path,
        combined_mask_path,
        cache_out_path,
        mask_fitments,
    )

    # Settle on the final output path for the cleaned image.
    # Check if outputting directly.
    if c_data.output_dir is not None:
        if c_data.output_dir.is_absolute():
            final_out_path = c_data.output_dir / original_img_path_as_png.name
        else:
            # Take the original image path, and place the image in a subdirectory.
            # This is for when multiple directories were passed in.
            final_out_path = (
                original_img_path_as_png.parent / c_data.output_dir / original_img_path_as_png.name
            )

        final_out_path.parent.mkdir(parents=True, exist_ok=True)
        final_cleaned_out_path = final_out_path.with_name(final_out_path.stem + "_clean.png")
        final_mask_out_path = final_out_path.with_name(final_out_path.stem + "_mask.png")

        logger.debug(f"Final output path: {final_cleaned_out_path}")

        # The arg parser should ensure that both can't be true at once, not like that'd be an issue, just plain silly.
        if not c_data.save_only_mask:
            # Save the final image.
            logger.debug(f"Saving final image to {final_cleaned_out_path}")
            base_image.save(final_cleaned_out_path)

        if not c_data.save_only_cleaned:
            # Save the final image.
            logger.debug(f"Saving final mask to {final_mask_out_path}")
            combined_mask.save(final_mask_out_path)

    return analytics


def save_denoising_data(
    original_img_path: Path,
    cleaned_path: Path,
    mask_path: Path,
    cache_path: Path,
    mask_fitments: list[st.MaskFittingResults],
):
    """
    Save the data needed for denoising.

    Gather the following to construct a MaskData struct:
    - original_path: Path
    - cleaned_path: Path
    - mask_path: Path
    - boxes_with_deviation: list[tuple[tuple[int, int, int, int], float]]

    :param original_img_path: The path to the original image.
    :param cleaned_path: The path to the cleaned image.
    :param mask_path: The path to the combined mask.
    :param cache_path: The path to the cache directory.
    :param mask_fitments: The mask fitments.
    """

    # Gather the data needed for denoising.
    boxes_with_deviation = [m.noise_mask_data for m in mask_fitments]

    # Save the data.
    mask_data = st.MaskData(original_img_path, cleaned_path, mask_path, boxes_with_deviation)
    mask_data.write_json(cache_path.with_name(cache_path.stem + "_mask_data.json"))



import json
import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import magic
from PIL import Image, ImageDraw

import pcleaner.config as cfg


class BoxType(Enum):
    BOX = 0
    EXTENDED_BOX = 1
    MERGED_EXT_BOX = 2
    REFERENCE_BOX = 3


@dataclass
class PageData:
    """
    This dataclass represents the json data generated by the ai model.
    It contains the image path, mask path, the boxes, the extended boxes,
    and the merged extended boxes.

    - boxes: The boxes generated by the ai model, only expanded slightly for extra padding.
      These are used to make a tight-fitting mask.
    - extended_boxes: The boxes expanded a lot and used for masking off potential false positives.
    - merged_extended_boxes: Overlapping extended boxes are merged into one box, to prevent
      conflicts of overlapping mask regions. The original extended boxes are still kept,
      to provide a more precise mask for the initial cut. These are used to cut out the mask when
      analyzing the fit.
    - reference_boxes: These are extensions of the merged extended boxes, to make sure the mask
      has room to grow in the base image when providing analysis. These are used to cut out the
      base image when analyzing the fit.

    Boxes are represented as tuples of (x1, y1, x2, y2), where (x1, y1) is the top left corner,
    and (x2, y2) is the bottom right corner.
    """

    image_path: str  # Path to the copied png.
    mask_path: str  # Path to the generated mask.png
    original_path: str  # Path to the original image. (used for relative output)
    boxes: list[tuple[int, int, int, int]]
    extended_boxes: list[tuple[int, int, int, int]]
    merged_extended_boxes: list[tuple[int, int, int, int]]
    reference_boxes: list[tuple[int, int, int, int]]
    _image_size: tuple[
        int, int
    ] = None  # Cache the image size, so we don't have to load the image every time.

    @classmethod
    def from_json(cls, json_str: str) -> "PageData":
        """
        Load a previously dumped PageData object from a json file.

        :param json_str: The json string to load from.
        """
        json_data = json.loads(json_str)
        return cls(
            json_data["image_path"],
            json_data["mask_path"],
            json_data["original_path"],
            json_data["boxes"],
            json_data["extended_boxes"],
            json_data["merged_extended_boxes"],
            json_data["reference_boxes"],
        )

    @property
    def image_size(self):
        if self._image_size is None:
            metadata = magic.from_file(self.image_path)
            size_str = re.search(r"(\d+) x (\d+)", metadata).groups()
            self._image_size = (int(size_str[0]), int(size_str[1]))
        return self._image_size

    def boxes_from_type(self, box_type: BoxType) -> list[tuple[int, int, int, int]]:
        match box_type:
            case BoxType.BOX:
                return self.boxes
            case BoxType.EXTENDED_BOX:
                return self.extended_boxes
            case BoxType.MERGED_EXT_BOX:
                return self.merged_extended_boxes
            case BoxType.REFERENCE_BOX:
                return self.reference_boxes
            case _:
                raise ValueError("Invalid box type.")

    def grow_boxes(self, padding: int, box_type: BoxType):
        """
        Uniformly grow all boxes by padding pixels.
        Checks that the boxes don't go out of bounds.

        :param padding: number of pixels to grow each box by.
        :param box_type: type of box to grow.
        """
        boxes = self.boxes_from_type(box_type)
        image_width, image_height = self.image_size

        for i in range(len(boxes)):
            x1, y1, x2, y2 = boxes[i]
            boxes[i] = (
                max(x1 - padding, 0),
                max(y1 - padding, 0),
                min(x2 + padding, image_width),
                min(y2 + padding, image_height),
            )

    def right_pad_boxes(self, padding: int, box_type: BoxType):
        """
        Right-pad all boxes by padding pixels.
        Checks that the boxes don't go out of bounds.

        :param padding: number of pixels to right-pad each box by.
        :param box_type: type of box to right-pad.
        """
        boxes = self.boxes_from_type(box_type)
        image_width, _ = self.image_size

        for i in range(len(boxes)):
            x1, y1, x2, y2 = boxes[i]
            boxes[i] = (x1, y1, min(x2 + padding, image_width), y2)

    @staticmethod
    def box_size(box: tuple[int, int, int, int]) -> int:
        """
        Calculate the size of a box.

        :param box: The box to calculate the size of.
        :return: The size of the box.
        """
        return (box[2] - box[0]) * (box[3] - box[1])

    def visualize(self, image_path: Path, out_dir: Path | None = None):
        """
        Visualize the boxes on an image.
        Typically, this would be used to check where on the original image the
        boxes are located.

        :param image_path: The path to the image to visualize the boxes on.
        :param out_dir: [Optional] The directory to save the visualization to.
            By default, the visualization is saved to the same directory as the image.
        """
        image = Image.open(image_path)
        draw = ImageDraw.Draw(image)
        for box in self.boxes:
            draw.rectangle(box, outline="green")
        for box in self.extended_boxes:
            draw.rectangle(box, outline="red")
        for box in self.merged_extended_boxes:
            draw.rectangle(box, outline="purple")
        for box in self.reference_boxes:
            draw.rectangle(box, outline="blue")
        # Save the image.
        out_path = image_path.with_stem(image_path.stem + "_boxes")
        if out_dir is not None:
            out_dir.mkdir(parents=True, exist_ok=True)
            out_path = out_dir / image_path.name
        image.save(out_path)

    def make_box_mask(self, image_size: tuple[int, int], box_type: BoxType) -> Image:
        """
        Draw the boxes as a bitmap mask, where 1 represents box, and 0 represents no box.
        This is in essence just an image where the color is either black or white.

        :param image_size: The size of the image to make the mask for.
        :param box_type: The type of box to use.
        :return: The mask. Image mode: "1"
        """
        box_mask = Image.new("1", image_size, (0,))
        draw = ImageDraw.Draw(box_mask)
        boxes = self.boxes_from_type(box_type)
        for box in boxes:
            draw.rectangle(box, fill=(1,))
        return box_mask

    def resolve_overlaps(self):
        """
        Copy the extended boxes to the merged extended boxes, and merge overlapping boxes.
        """

        def boxes_overlap(box1: tuple[int, int, int, int], box2: tuple[int, int, int, int]) -> bool:
            x1_min, y1_min, x1_max, y1_max = box1
            x2_min, y2_min, x2_max, y2_max = box2

            x_overlap = (x1_min <= x2_max) and (x2_min <= x1_max)
            y_overlap = (y1_min <= y2_max) and (y2_min <= y1_max)

            # If both dimensions overlap, then the boxes overlap.
            return x_overlap and y_overlap

        def merge_boxes(
            box1: tuple[int, int, int, int], box2: tuple[int, int, int, int]
        ) -> tuple[int, int, int, int]:
            x1_min, y1_min, x1_max, y1_max = box1
            x2_min, y2_min, x2_max, y2_max = box2

            x_min = min(x1_min, x2_min)
            y_min = min(y1_min, y2_min)
            x_max = max(x1_max, x2_max)
            y_max = max(y1_max, y2_max)

            return x_min, y_min, x_max, y_max

        # Place the extended boxes in the merged extended boxes, merging overlapping boxes.
        merge_queue = set(self.extended_boxes)
        merged_boxes = []
        while merge_queue:
            box = merge_queue.pop()
            # Find all boxes that overlap with this box.
            overlapping_boxes = [b for b in merge_queue if boxes_overlap(box, b)]
            # Merge all overlapping boxes.
            for b in overlapping_boxes:
                box = merge_boxes(box, b)
                merge_queue.remove(b)
            merged_boxes.append(box)

        self.merged_extended_boxes = merged_boxes


@dataclass
class MaskFittingResults:
    """
    This is a simple struct to hold the results from the mask fitting process.
    Since it returns a lot of data, this is more readable.
    """

    best_mask: Image
    median_color: int
    mask_coords: tuple[int, int]
    analytics_page_path: Path
    analytics_std_deviation: float
    analytics_mask_index: int
    mask_box: tuple[int, int, int, int]  # Used for the denoising process.
    debug_masks: list[Image]

    @property
    def analytics(self) -> tuple[Path, bool, int, float]:
        return (
            self.analytics_page_path,
            self.best_mask is not None,
            self.analytics_mask_index,
            self.analytics_std_deviation,
        )

    @property
    def failed(self) -> bool:
        return self.best_mask is None

    @property
    def mask_data(self) -> tuple[Image, int, tuple[int, int]]:
        return self.best_mask, self.median_color, self.mask_coords

    @property
    def noise_mask_data(self) -> tuple[tuple[int, int, int, int], float]:
        return self.mask_box, self.analytics_std_deviation


@dataclass
class CleanerData:
    """
    This is a simple struct to hold the inputs for the cleaner.
    The data is a tuple of:
    - The json file path.
    - The image output directory. When none, output to the cache dir for denoising to continue.
    - The image cache directory.
    - The save only mask flag.
    - The save only cleaned flag.
    - The show masks flag. (when true, save intermediate masks to the cache directory)
    - The debug flag.
    """

    json_path: Path
    output_dir: Path | None
    cache_dir: Path
    cleaner_config: cfg.CleanerConfig
    save_only_mask: bool
    save_only_cleaned: bool
    show_masks: bool
    debug: bool


@dataclass
class MaskData:
    """
    This is a simple struct to hold all the extra information needed to perform the
    denoising process.

    - The original image path.
    - The cleaned image path.
    - The mask image path.
    - The box coordinates with their respective standard deviation for the masks.
    """

    original_path: Path
    cleaned_path: Path
    mask_path: Path
    boxes_with_deviation: list[tuple[tuple[int, int, int, int], float]]

    @classmethod
    def from_json(cls, json_str: str) -> "MaskData":
        """
        Create a MaskData object from a json string.

        :param json_str: The json string.
        :return: The MaskData object.
        """
        data = json.loads(json_str)
        return cls(
            Path(data["original_path"]),
            Path(data["cleaned_path"]),
            Path(data["mask_path"]),
            data["boxes_with_deviation"],
        )

    def write_json(self, json_path: Path):
        """
        Write the MaskData object to a json file.

        :param json_path: The path to write the json file to.
        """
        # Convert the Path objects to strings.
        data = {
            "original_path": str(self.original_path),
            "cleaned_path": str(self.cleaned_path),
            "mask_path": str(self.mask_path),
            "boxes_with_deviation": self.boxes_with_deviation,
        }
        with open(json_path, "w") as f:
            json.dump(data, f, indent=4)


@dataclass
class DenoiserData:
    """
    This is a simple struct to hold the inputs for the denoiser.
    The data is a tuple of:
    - The json file path.
    - The image output directory.
    - The image cache directory.
    - The denoiser config.
    - The save only mask flag.
    - The save only cleaned flag.
    - The separate noise mask flag.
    - The show masks flag. (when true, save intermediate masks to the cache directory)
    - The debug flag.
    """

    json_path: Path
    output_dir: Path
    cache_dir: Path
    denoiser_config: cfg.DenoiserConfig
    save_only_mask: bool
    save_only_cleaned: bool
    separate_noise_masks: bool
    show_masks: bool
    debug: bool


@dataclass
class DenoiseAnalytic:
    """
    Analytics data to visualize the denoising performance.
    - The standard deviations of the mask selection process. They are shown here,
      due to them being relevant to the min-threshold for choosing masks to denoise.
    """

    std_deviations: tuple[float, ...]



from copy import deepcopy
from pathlib import Path
from ...utils.yolov5_utils import *


from .common import *


class Detect(nn.Module):
    stride = None  # strides computed during build
    onnx_dynamic = False  # ONNX export parameter

    def __init__(self, nc=80, anchors=(), ch=(), inplace=True):  # detection layer
        super().__init__()
        self.nc = nc  # number of classes
        self.no = nc + 5  # number of outputs per anchor
        self.nl = len(anchors)  # number of detection layers
        self.na = len(anchors[0]) // 2  # number of anchors
        self.grid = [torch.zeros(1)] * self.nl  # init grid
        self.anchor_grid = [torch.zeros(1)] * self.nl  # init anchor grid
        self.register_buffer(
            "anchors", torch.tensor(anchors).float().view(self.nl, -1, 2)
        )  # shape(nl,na,2)
        self.m = nn.ModuleList(nn.Conv2d(x, self.no * self.na, 1) for x in ch)  # output conv
        self.inplace = inplace  # use in-place ops (e.g. slice assignment)

    def forward(self, x):
        z = []  # inference output
        for i in range(self.nl):
            x[i] = self.m[i](x[i])  # conv
            bs, _, ny, nx = x[i].shape  # x(bs,255,20,20) to x(bs,3,20,20,85)
            x[i] = x[i].view(bs, self.na, self.no, ny, nx).permute(0, 1, 3, 4, 2).contiguous()

            if not self.training:  # inference
                if self.onnx_dynamic or self.grid[i].shape[2:4] != x[i].shape[2:4]:
                    self.grid[i], self.anchor_grid[i] = self._make_grid(nx, ny, i)

                y = x[i].sigmoid()
                if self.inplace:
                    y[..., 0:2] = (y[..., 0:2] * 2 - 0.5 + self.grid[i]) * self.stride[i]  # xy
                    y[..., 2:4] = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i]  # wh
                else:  # for YOLOv5 on AWS Inferentia https://github.com/ultralytics/yolov5/pull/2953
                    xy = (y[..., 0:2] * 2 - 0.5 + self.grid[i]) * self.stride[i]  # xy
                    wh = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i]  # wh
                    y = torch.cat((xy, wh, y[..., 4:]), -1)
                z.append(y.view(bs, -1, self.no))

        return x if self.training else (torch.cat(z, 1), x)

    def _make_grid(self, nx=20, ny=20, i=0):
        d = self.anchors[i].device
        if check_version(
            torch.__version__, "1.10.0"
        ):  # torch>=1.10.0 meshgrid workaround for torch>=0.7 compatibility
            yv, xv = torch.meshgrid(
                [torch.arange(ny, device=d), torch.arange(nx, device=d)], indexing="ij"
            )
        else:
            yv, xv = torch.meshgrid([torch.arange(ny, device=d), torch.arange(nx, device=d)])
        grid = torch.stack((xv, yv), 2).expand((1, self.na, ny, nx, 2)).float()
        anchor_grid = (
            (self.anchors[i].clone() * self.stride[i])
            .view((1, self.na, 1, 1, 2))
            .expand((1, self.na, ny, nx, 2))
            .float()
        )
        return grid, anchor_grid


class Model(nn.Module):
    def __init__(
        self, cfg="yolov5s.yaml", ch=3, nc=None, anchors=None
    ):  # model, input channels, number of classes
        super().__init__()
        self.out_indices = None
        if isinstance(cfg, dict):
            self.yaml = cfg  # model dict
        else:  # is *.yaml
            import yaml  # for torch hub

            self.yaml_file = Path(cfg).name
            with open(cfg, encoding="ascii", errors="ignore") as f:
                self.yaml = yaml.safe_load(f)  # model dict

        # Define model
        ch = self.yaml["ch"] = self.yaml.get("ch", ch)  # input channels
        if nc and nc != self.yaml["nc"]:
            # LOGGER.info(f"Overriding model.yaml nc={self.yaml['nc']} with nc={nc}")
            self.yaml["nc"] = nc  # override yaml value
        if anchors:
            # LOGGER.info(f'Overriding model.yaml anchors with anchors={anchors}')
            self.yaml["anchors"] = round(anchors)  # override yaml value
        self.model, self.save = parse_model(deepcopy(self.yaml), ch=[ch])  # model, savelist
        self.names = [str(i) for i in range(self.yaml["nc"])]  # default names
        self.inplace = self.yaml.get("inplace", True)

        # Build strides, anchors
        m = self.model[-1]  # Detect()
        # with torch.no_grad():
        if isinstance(m, Detect):
            s = 256  # 2x min stride
            m.inplace = self.inplace
            m.stride = torch.tensor(
                [s / x.shape[-2] for x in self.forward(torch.zeros(1, ch, s, s))]
            )  # forward
            m.anchors /= m.stride.view(-1, 1, 1)
            check_anchor_order(m)
            self.stride = m.stride
            self._initialize_biases()  # only run once

        # Init weights, biases
        initialize_weights(self)

    def forward(self, x, augment=False, profile=False, visualize=False, detect=False):
        if augment:
            return self._forward_augment(x)  # augmented inference, None
        return self._forward_once(
            x, profile, visualize, detect=detect
        )  # single-scale inference, train

    def _forward_augment(self, x):
        img_size = x.shape[-2:]  # height, width
        s = [1, 0.83, 0.67]  # scales
        f = [None, 3, None]  # flips (2-ud, 3-lr)
        y = []  # outputs
        for si, fi in zip(s, f):
            xi = scale_img(x.flip(fi) if fi else x, si, gs=int(self.stride.max()))
            yi = self._forward_once(xi)[0]  # forward
            # cv2.imwrite(f'img_{si}.jpg', 255 * xi[0].cpu().numpy().transpose((1, 2, 0))[:, :, ::-1])  # save
            yi = self._descale_pred(yi, fi, si, img_size)
            y.append(yi)
        y = self._clip_augmented(y)  # clip augmented tails
        return torch.cat(y, 1), None  # augmented inference, train

    def _forward_once(self, x, profile=False, visualize=False, detect=False):
        y, dt = [], []  # outputs
        z = []
        for ii, m in enumerate(self.model):
            if m.f != -1:  # if not from previous layer
                x = (
                    y[m.f] if isinstance(m.f, int) else [x if j == -1 else y[j] for j in m.f]
                )  # from earlier layers
            if profile:
                self._profile_one_layer(m, x, dt)
            x = m(x)  # run
            y.append(x if m.i in self.save else None)  # save output
            if self.out_indices is not None:
                if m.i in self.out_indices:
                    z.append(x)
        if self.out_indices is not None:
            if detect:
                return x, z
            else:
                return z
        else:
            return x

    def _descale_pred(self, p, flips, scale, img_size):
        # de-scale predictions following augmented inference (inverse operation)
        if self.inplace:
            p[..., :4] /= scale  # de-scale
            if flips == 2:
                p[..., 1] = img_size[0] - p[..., 1]  # de-flip ud
            elif flips == 3:
                p[..., 0] = img_size[1] - p[..., 0]  # de-flip lr
        else:
            x, y, wh = p[..., 0:1] / scale, p[..., 1:2] / scale, p[..., 2:4] / scale  # de-scale
            if flips == 2:
                y = img_size[0] - y  # de-flip ud
            elif flips == 3:
                x = img_size[1] - x  # de-flip lr
            p = torch.cat((x, y, wh, p[..., 4:]), -1)
        return p

    def _clip_augmented(self, y):
        # Clip YOLOv5 augmented inference tails
        nl = self.model[-1].nl  # number of detection layers (P3-P5)
        g = sum(4**x for x in range(nl))  # grid points
        e = 1  # exclude layer count
        i = (y[0].shape[1] // g) * sum(4**x for x in range(e))  # indices
        y[0] = y[0][:, :-i]  # large
        i = (y[-1].shape[1] // g) * sum(4 ** (nl - 1 - x) for x in range(e))  # indices
        y[-1] = y[-1][:, i:]  # small
        return y

    def _profile_one_layer(self, m, x, dt):
        c = isinstance(m, Detect)  # is final layer, copy input as inplace fix
        for _ in range(10):
            m(x.copy() if c else x)

    def _initialize_biases(self, cf=None):  # initialize biases into Detect(), cf is class frequency
        # https://arxiv.org/abs/1708.02002 section 3.3
        # cf = torch.bincount(torch.tensor(np.concatenate(dataset.labels, 0)[:, 0]).long(), minlength=nc) + 1.
        m = self.model[-1]  # Detect() module
        for mi, s in zip(m.m, m.stride):  # from
            b = mi.bias.view(m.na, -1)  # conv.bias(255) to (3,85)
            b.data[:, 4] += math.log(8 / (640 / s) ** 2)  # obj (8 objects per 640 image)
            b.data[:, 5:] += (
                math.log(0.6 / (m.nc - 0.999999)) if cf is None else torch.log(cf / cf.sum())
            )  # cls
            mi.bias = torch.nn.Parameter(b.view(-1), requires_grad=True)

    def _print_biases(self):
        m = self.model[-1]  # Detect() module
        for mi in m.m:  # from
            b = mi.bias.detach().view(m.na, -1).T  # conv.bias(255) to (3,85)

    def fuse(self):  # fuse model Conv2d() + BatchNorm2d() layers
        for m in self.model.modules():
            if isinstance(m, (Conv, DWConv)) and hasattr(m, "bn"):
                m.conv = fuse_conv_and_bn(m.conv, m.bn)  # update conv
                delattr(m, "bn")  # remove batchnorm
                m.forward = m.forward_fuse  # update forward
        # self.info()
        return self

    # def info(self, verbose=False, img_size=640):  # print model information
    #     model_info(self, verbose, img_size)

    def _apply(self, fn):
        # Apply to(), cpu(), cuda(), half() to model tensors that are not parameters or registered buffers
        self = super()._apply(fn)
        m = self.model[-1]  # Detect()
        if isinstance(m, Detect):
            m.stride = fn(m.stride)
            m.grid = list(map(fn, m.grid))
            if isinstance(m.anchor_grid, list):
                m.anchor_grid = list(map(fn, m.anchor_grid))
        return self


def parse_model(d, ch):  # model_dict, input_channels(3)
    # LOGGER.info(f"\n{'':>3}{'from':>18}{'n':>3}{'params':>10}  {'module':<40}{'arguments':<30}")
    anchors, nc, gd, gw = d["anchors"], d["nc"], d["depth_multiple"], d["width_multiple"]
    na = (len(anchors[0]) // 2) if isinstance(anchors, list) else anchors  # number of anchors
    no = na * (nc + 5)  # number of outputs = anchors * (classes + 5)

    layers, save, c2 = [], [], ch[-1]  # layers, savelist, ch out
    for i, (f, n, m, args) in enumerate(d["backbone"] + d["head"]):  # from, number, module, args
        m = eval(m) if isinstance(m, str) else m  # eval strings
        for j, a in enumerate(args):
            try:
                args[j] = eval(a) if isinstance(a, str) else a  # eval strings
            except NameError:
                pass

        n = n_ = max(round(n * gd), 1) if n > 1 else n  # depth gain
        if m in [
            Conv,
            GhostConv,
            Bottleneck,
            GhostBottleneck,
            SPP,
            SPPF,
            DWConv,
            Focus,
            BottleneckCSP,
            C3,
            C3TR,
            C3SPP,
            C3Ghost,
        ]:
            c1, c2 = ch[f], args[0]
            if c2 != no:  # if not output
                c2 = make_divisible(c2 * gw, 8)

            args = [c1, c2, *args[1:]]
            if m in [BottleneckCSP, C3, C3TR, C3Ghost]:
                args.insert(2, n)  # number of repeats
                n = 1
        elif m is nn.BatchNorm2d:
            args = [ch[f]]
        elif m is Concat:
            c2 = sum(ch[x] for x in f)
        elif m is Detect:
            args.append([ch[x] for x in f])
            if isinstance(args[1], int):  # number of anchors
                args[1] = [list(range(args[1] * 2))] * len(f)
        elif m is Contract:
            c2 = ch[f] * args[0] ** 2
        elif m is Expand:
            c2 = ch[f] // args[0] ** 2
        else:
            c2 = ch[f]

        m_ = nn.Sequential(*(m(*args) for _ in range(n))) if n > 1 else m(*args)  # module
        t = str(m)[8:-2].replace("__main__.", "")  # module type
        np = sum(x.numel() for x in m_.parameters())  # number params
        m_.i, m_.f, m_.type, m_.np = i, f, t, np  # attach index, 'from' index, type, number params
        # LOGGER.info(f'{i:>3}{str(f):>18}{n_:>3}{np:10.0f}  {t:<40}{str(args):<30}')  # print
        save.extend(
            x % i for x in ([f] if isinstance(f, int) else f) if x != -1
        )  # append to savelist
        layers.append(m_)
        if i == 0:
            ch = []
        ch.append(c2)
    return nn.Sequential(*layers), sorted(save)


def load_yolov5(weights, map_location="cuda", fuse=True, inplace=True, out_indices=[1, 3, 5, 7, 9]):
    if isinstance(weights, str):
        ckpt = torch.load(weights, map_location=map_location)  # load
    else:
        ckpt = weights

    if fuse:
        model = ckpt["model"].float().fuse().eval()  # FP32 model
    else:
        model = ckpt["model"].float().eval()  # without layer fuse

    # Compatibility updates
    for m in model.modules():
        if type(m) in [nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU, Detect, Model]:
            m.inplace = inplace  # pytorch 1.7.0 compatibility
            if type(m) is Detect:
                if not isinstance(m.anchor_grid, list):  # new Detect Layer compatibility
                    delattr(m, "anchor_grid")
                    setattr(m, "anchor_grid", [torch.zeros(1)] * m.nl)
        elif type(m) is Conv:
            m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility
    model.out_indices = out_indices
    return model


@torch.no_grad()
def load_yolov5_ckpt(
    weights, map_location="cpu", fuse=True, inplace=True, out_indices=[1, 3, 5, 7, 9]
):
    if isinstance(weights, str):
        ckpt = torch.load(weights, map_location=map_location)  # load
    else:
        ckpt = weights

    model = Model(ckpt["cfg"])
    model.load_state_dict(ckpt["weights"], strict=True)

    if fuse:
        model = model.float().fuse().eval()  # FP32 model
    else:
        model = model.float().eval()  # without layer fuse

    # Compatibility updates
    for m in model.modules():
        if type(m) in [nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU, Detect, Model]:
            m.inplace = inplace  # pytorch 1.7.0 compatibility
            if type(m) is Detect:
                if not isinstance(m.anchor_grid, list):  # new Detect Layer compatibility
                    delattr(m, "anchor_grid")
                    setattr(m, "anchor_grid", [torch.zeros(1)] * m.nl)
        elif type(m) is Conv:
            m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility
    model.out_indices = out_indices
    return model



import copy

import cv2
import torch
import torch.nn as nn

from .models.yolov5.common import C3, Conv
from .models.yolov5.yolo import load_yolov5_ckpt
from .utils.weight_init import init_weights
from .utils.yolov5_utils import fuse_conv_and_bn

TEXTDET_MASK = 0
TEXTDET_DET = 1
TEXTDET_INFERENCE = 2


class double_conv_up_c3(nn.Module):
    def __init__(self, in_ch, mid_ch, out_ch, act=True):
        super(double_conv_up_c3, self).__init__()
        self.conv = nn.Sequential(
            C3(in_ch + mid_ch, mid_ch, act=act),
            nn.ConvTranspose2d(mid_ch, out_ch, kernel_size=4, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.conv(x)


class double_conv_c3(nn.Module):
    def __init__(self, in_ch, out_ch, stride=1, act=True):
        super(double_conv_c3, self).__init__()
        if stride > 1:
            self.down = nn.AvgPool2d(2, stride=2) if stride > 1 else None
        self.conv = C3(in_ch, out_ch, act=act)

    def forward(self, x):
        if self.down is not None:
            x = self.down(x)
        x = self.conv(x)
        return x


class UnetHead(nn.Module):
    def __init__(self, act=True) -> None:

        super(UnetHead, self).__init__()
        self.down_conv1 = double_conv_c3(512, 512, 2, act=act)
        self.upconv0 = double_conv_up_c3(0, 512, 256, act=act)
        self.upconv2 = double_conv_up_c3(256, 512, 256, act=act)
        self.upconv3 = double_conv_up_c3(0, 512, 256, act=act)
        self.upconv4 = double_conv_up_c3(128, 256, 128, act=act)
        self.upconv5 = double_conv_up_c3(64, 128, 64, act=act)
        self.upconv6 = nn.Sequential(
            nn.ConvTranspose2d(64, 1, kernel_size=4, stride=2, padding=1, bias=False), nn.Sigmoid()
        )

    def forward(self, f160, f80, f40, f20, f3, forward_mode=TEXTDET_MASK):
        # input: 640@3
        d10 = self.down_conv1(f3)  # 512@10
        u20 = self.upconv0(d10)  # 256@10
        u40 = self.upconv2(torch.cat([f20, u20], dim=1))  # 256@40

        if forward_mode == TEXTDET_DET:
            return f80, f40, u40
        else:
            u80 = self.upconv3(torch.cat([f40, u40], dim=1))  # 256@80
            u160 = self.upconv4(torch.cat([f80, u80], dim=1))  # 128@160
            u320 = self.upconv5(torch.cat([f160, u160], dim=1))  # 64@320
            mask = self.upconv6(u320)
            if forward_mode == TEXTDET_MASK:
                return mask
            else:
                return mask, [f80, f40, u40]

    def init_weight(self, init_func):
        self.apply(init_func)


class DBHead(nn.Module):
    def __init__(self, in_channels, k=50, shrink_with_sigmoid=True, act=True):
        super().__init__()
        self.k = k
        self.shrink_with_sigmoid = shrink_with_sigmoid
        self.upconv3 = double_conv_up_c3(0, 512, 256, act=act)
        self.upconv4 = double_conv_up_c3(128, 256, 128, act=act)
        self.conv = nn.Sequential(
            nn.Conv2d(128, in_channels, 1), nn.BatchNorm2d(in_channels), nn.ReLU(inplace=True)
        )
        self.binarize = nn.Sequential(
            nn.Conv2d(in_channels, in_channels // 4, 3, padding=1),
            nn.BatchNorm2d(in_channels // 4),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels // 4, in_channels // 4, 2, 2),
            nn.BatchNorm2d(in_channels // 4),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels // 4, 1, 2, 2),
        )
        self.thresh = self._init_thresh(in_channels)

    def forward(self, f80, f40, u40, shrink_with_sigmoid=True, step_eval=False):
        shrink_with_sigmoid = self.shrink_with_sigmoid
        u80 = self.upconv3(torch.cat([f40, u40], dim=1))  # 256@80
        x = self.upconv4(torch.cat([f80, u80], dim=1))  # 128@160
        x = self.conv(x)
        threshold_maps = self.thresh(x)
        x = self.binarize(x)
        shrink_maps = torch.sigmoid(x)

        if self.training:
            binary_maps = self.step_function(shrink_maps, threshold_maps)
            if shrink_with_sigmoid:
                return torch.cat((shrink_maps, threshold_maps, binary_maps), dim=1)
            else:
                return torch.cat((shrink_maps, threshold_maps, binary_maps, x), dim=1)
        else:
            if step_eval:
                return self.step_function(shrink_maps, threshold_maps)
            else:
                return torch.cat((shrink_maps, threshold_maps), dim=1)

    def init_weight(self, init_func):
        self.apply(init_func)

    def _init_thresh(self, inner_channels, serial=False, smooth=False, bias=False):
        in_channels = inner_channels
        if serial:
            in_channels += 1
        self.thresh = nn.Sequential(
            nn.Conv2d(in_channels, inner_channels // 4, 3, padding=1, bias=bias),
            nn.BatchNorm2d(inner_channels // 4),
            nn.ReLU(inplace=True),
            self._init_upsample(inner_channels // 4, inner_channels // 4, smooth=smooth, bias=bias),
            nn.BatchNorm2d(inner_channels // 4),
            nn.ReLU(inplace=True),
            self._init_upsample(inner_channels // 4, 1, smooth=smooth, bias=bias),
            nn.Sigmoid(),
        )
        return self.thresh

    def _init_upsample(self, in_channels, out_channels, smooth=False, bias=False):
        if smooth:
            inter_out_channels = out_channels
            if out_channels == 1:
                inter_out_channels = in_channels
            module_list = [
                nn.Upsample(scale_factor=2, mode="nearest"),
                nn.Conv2d(in_channels, inter_out_channels, 3, 1, 1, bias=bias),
            ]
            if out_channels == 1:
                module_list.append(
                    nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=1, bias=True)
                )
            return nn.Sequential(module_list)
        else:
            return nn.ConvTranspose2d(in_channels, out_channels, 2, 2)

    def step_function(self, x, y):
        return torch.reciprocal(1 + torch.exp(-self.k * (x - y)))


class TextDetector(nn.Module):
    def __init__(self, weights, map_location="cpu", forward_mode=TEXTDET_MASK, act=True):
        super(TextDetector, self).__init__()

        yolov5s_backbone = load_yolov5_ckpt(weights=weights, map_location=map_location)
        yolov5s_backbone.eval()
        out_indices = [1, 3, 5, 7, 9]
        yolov5s_backbone.out_indices = out_indices
        yolov5s_backbone.model = yolov5s_backbone.model[: max(out_indices) + 1]
        self.act = act
        self.seg_net = UnetHead(act=act)
        self.backbone = yolov5s_backbone
        self.dbnet = None
        self.forward_mode = forward_mode

    def train_mask(self):
        self.forward_mode = TEXTDET_MASK
        self.backbone.eval()
        self.seg_net.train()

    def initialize_db(self, unet_weights):
        self.dbnet = DBHead(64, act=self.act)
        self.seg_net.load_state_dict(torch.load(unet_weights, map_location="cpu")["weights"])
        self.dbnet.init_weight(init_weights)
        self.dbnet.upconv3 = copy.deepcopy(self.seg_net.upconv3)
        self.dbnet.upconv4 = copy.deepcopy(self.seg_net.upconv4)
        del self.seg_net.upconv3
        del self.seg_net.upconv4
        del self.seg_net.upconv5
        del self.seg_net.upconv6
        # del self.seg_net.conv_mask

    def train_db(self):
        self.forward_mode = TEXTDET_DET
        self.backbone.eval()
        self.seg_net.eval()
        self.dbnet.train()

    def forward(self, x):
        forward_mode = self.forward_mode
        with torch.no_grad():
            outs = self.backbone(x)
        if forward_mode == TEXTDET_MASK:
            return self.seg_net(*outs, forward_mode=forward_mode)
        elif forward_mode == TEXTDET_DET:
            with torch.no_grad():
                outs = self.seg_net(*outs, forward_mode=forward_mode)
            return self.dbnet(*outs)


def get_base_det_models(model_path, device="cpu", half=False, act="leaky"):
    textdetector_dict = torch.load(model_path, map_location=device)
    blk_det = load_yolov5_ckpt(textdetector_dict["blk_det"], map_location=device)
    text_seg = UnetHead(act=act)
    text_seg.load_state_dict(textdetector_dict["text_seg"])
    text_det = DBHead(64, act=act)
    text_det.load_state_dict(textdetector_dict["text_det"])
    if half:
        return blk_det.eval().half(), text_seg.eval().half(), text_det.eval().half()
    return blk_det.eval().to(device), text_seg.eval().to(device), text_det.eval().to(device)


class TextDetBase(nn.Module):
    def __init__(self, model_path, device="cpu", half=False, fuse=False, act="leaky"):
        super(TextDetBase, self).__init__()
        self.blk_det, self.text_seg, self.text_det = get_base_det_models(model_path, device, half, act=act)
        if fuse:
            self.fuse()

    def fuse(self):
        def _fuse(model):
            for m in model.modules():
                if isinstance(m, (Conv)) and hasattr(m, "bn"):
                    m.conv = fuse_conv_and_bn(m.conv, m.bn)  # update conv
                    delattr(m, "bn")  # remove batchnorm
                    m.forward = m.forward_fuse  # update forward
            return model

        self.text_seg = _fuse(self.text_seg)
        self.text_det = _fuse(self.text_det)

    def forward(self, features):
        blks, features = self.blk_det(features, detect=True)
        mask, features = self.text_seg(*features, forward_mode=TEXTDET_INFERENCE)
        lines = self.text_det(*features, step_eval=False)
        return blks[0], mask, lines


class TextDetBaseDNN:
    def __init__(self, input_size, model_path):
        self.input_size = input_size
        self.model = cv2.dnn.readNetFromONNX(model_path)
        self.uoln = self.model.getUnconnectedOutLayersNames()

    def __call__(self, im_in):
        blob = cv2.dnn.blobFromImage(im_in, scalefactor=1 / 255.0, size=(self.input_size, self.input_size))
        self.model.setInput(blob)
        blks, mask, lines_map = self.model.forward(self.uoln)
        return blks, mask, lines_map



# import os
# import logging
# import wandb
# import torch
#
#
# def set_logging(name=None, verbose=True):
#     for handler in logging.root.handlers[:]:
#         logging.root.removeHandler(handler)
#     # Sets level and returns logger
#     rank = int(os.getenv("RANK", -1))  # rank in world for Multi-GPU trainings
#     logging.basicConfig(
#         format="%(message)s", level=logging.INFO if (verbose and rank in (-1, 0)) else logging.WARNING
#     )
#     return logging.getLogger(name)
#
#
# LOGGER = set_logging(__name__)  # define globally (used in train.py, val.py, detect.py, etc.)
#
# LOGGERS = ("csv", "tb", "wandb")
#
# CUDA = True if torch.cuda.is_available() else False
# DEVICE = "cuda" if CUDA else "cpu"
#
# LOGGER_WANDB = "wandb"
# LOGGER_TENSORBOARD = "tb"
#
#
# class Loggers:
#     def __init__(self, hyp):
#         self.type = hyp["logger"]["type"]
#         self.epochs = hyp["train"]["epochs"]
#         self.wandb = None
#         self.writer = None
#         if self.type == LOGGER_WANDB:
#             if hyp["logger"]["project"] == "":
#                 project = "ComicTextDetector"
#             else:
#                 project = hyp["logger"]["project"]
#             if hyp["logger"]["run_id"] == "":
#                 self.wandb = wandb.init(project=project, config=hyp, resume="allow")
#             else:
#                 self.wandb = wandb.init(
#                     project=project, config=hyp, resume="must", id=hyp["logger"]["run_id"]
#                 )
#         elif self.type == LOGGER_TENSORBOARD:
#             from torch.utils.tensorboard import SummaryWriter
#
#             self.writer = SummaryWriter(hyp["data"]["save_dir"])
#
#     def on_train_batch_end(self, metrics):
#         # Callback runs on train batch end
#         if self.wandb:
#             self.wandb.log(metrics)
#         pass
#
#     def on_train_epoch_end(self, epoch, metrics):
#         LOGGER.info(f"fin epoch {epoch}/{self.epochs}, metrics: {metrics}")
#         if self.type == LOGGER_WANDB:
#             self.wandb.log(metrics)
#         elif self.type == LOGGER_TENSORBOARD:
#             for key in metrics.keys():
#                 self.writer.add_scalar(key, metrics[key], epoch)
#
#     def on_model_save(self, last, epoch, final_epoch, best_fitness, fi):
#         # Callback runs on model save event
#         if self.wandb:
#             if ((epoch + 1) % self.opt.save_period == 0 and not final_epoch) and self.opt.save_period != -1:
#                 self.wandb.log_model(last.parent, self.opt, epoch, fi, best_model=best_fitness == fi)



import json
import os.path as osp
from pathlib import Path
from typing import Union

import cv2
import numpy as np
import torch
from tqdm import tqdm

from .basemodel import TextDetBase, TextDetBaseDNN
from .utils.db_utils import SegDetectorRepresenter
from .utils.imgproc_utils import letterbox, xyxy2yolo, get_yololabel_strings
from .utils.io_utils import imread, imwrite, find_all_imgs, NumpyEncoder
from .utils.textblock import TextBlock, group_output, visualize_textblocks
from .utils.textmask import (
    refine_mask,
    refine_undetected_mask,
    REFINEMASK_INPAINT,
    REFINEMASK_ANNOTATION,
)
from .utils.yolov5_utils import non_max_suppression


def model2annotations(model_path, img_dir_list, save_dir, save_json=False):
    if isinstance(img_dir_list, str):
        img_dir_list = [img_dir_list]
    cuda = torch.cuda.is_available()
    device = "cuda" if cuda else "cpu"
    model = TextDetector(model_path=model_path, input_size=1024, device=device, act="leaky")
    imglist = []
    for img_dir in img_dir_list:
        imglist += find_all_imgs(img_dir, abs_path=True)
    for img_path in tqdm(imglist):
        imgname = osp.basename(img_path)
        img = imread(img_path)
        im_h, im_w = img.shape[:2]
        imname = imgname.replace(Path(imgname).suffix, "")
        maskname = "mask-" + imname + ".png"
        poly_save_path = osp.join(save_dir, "line-" + imname + ".txt")
        mask, mask_refined, blk_list = model(
            img, refine_mode=REFINEMASK_ANNOTATION, keep_undetected_mask=True
        )
        polys = []
        blk_xyxy = []
        blk_dict_list = []
        for blk in blk_list:
            polys += blk.lines
            blk_xyxy.append(blk.xyxy)
            blk_dict_list.append(blk.to_dict())
        blk_xyxy = xyxy2yolo(blk_xyxy, im_w, im_h)
        if blk_xyxy is not None:
            cls_list = [1] * len(blk_xyxy)
            yolo_label = get_yololabel_strings(cls_list, blk_xyxy)
        else:
            yolo_label = ""
        with open(osp.join(save_dir, imname + ".txt"), "w", encoding="utf8") as f:
            f.write(yolo_label)

        # num_labels, labels, stats, centroids = cv2.connectedComponentsWithStats(mask)
        # _, mask = cv2.threshold(mask, 50, 255, cv2.THRESH_BINARY)
        # draw_connected_labels(num_labels, labels, stats, centroids)
        # visualize_textblocks(img, blk_list)
        # cv2.imshow('rst', img)
        # cv2.imshow('mask', mask)
        # cv2.imshow('mask_refined', mask_refined)
        # cv2.waitKey(0)

        if len(polys) != 0:
            if isinstance(polys, list):
                polys = np.array(polys)
            polys = polys.reshape(-1, 8)
            np.savetxt(poly_save_path, polys, fmt="%d")
        if save_json:
            with open(osp.join(save_dir, imname + ".json"), "w", encoding="utf8") as f:
                f.write(json.dumps(blk_dict_list, ensure_ascii=False, cls=NumpyEncoder))
        imwrite(osp.join(save_dir, imgname), img)
        imwrite(osp.join(save_dir, maskname), mask_refined)


def preprocess_img(
    img, input_size=(1024, 1024), device="cpu", bgr2rgb=True, half=False, to_tensor=True
):
    if bgr2rgb:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img_in, ratio, (dw, dh) = letterbox(img, new_shape=input_size, auto=False, stride=64)
    if to_tensor:
        img_in = img_in.transpose((2, 0, 1))[::-1]  # HWC to CHW, BGR to RGB
        img_in = np.array([np.ascontiguousarray(img_in)]).astype(np.float32) / 255
        if to_tensor:
            img_in = torch.from_numpy(img_in).to(device)
            if half:
                img_in = img_in.half()
    return img_in, ratio, int(dw), int(dh)


def postprocess_mask(img: Union[torch.Tensor, np.ndarray], thresh=None):
    # img = img.permute(1, 2, 0)
    if isinstance(img, torch.Tensor):
        img = img.squeeze_()
        if img.device != "cpu":
            img = img.detach_().cpu()
        img = img.numpy()
    else:
        img = img.squeeze()
    if thresh is not None:
        img = img > thresh
    img = img * 255
    # if isinstance(img, torch.Tensor):

    return img.astype(np.uint8)


def postprocess_yolo(det, conf_thresh, nms_thresh, resize_ratio, sort_func=None):
    det = non_max_suppression(det, conf_thresh, nms_thresh)[0]
    # bbox = det[..., 0:4]
    if det.device != "cpu":
        det = det.detach_().cpu().numpy()
    det[..., [0, 2]] = det[..., [0, 2]] * resize_ratio[0]
    det[..., [1, 3]] = det[..., [1, 3]] * resize_ratio[1]
    if sort_func is not None:
        det = sort_func(det)

    blines = det[..., 0:4].astype(np.int32)
    confs = np.round(det[..., 4], 3)
    cls = det[..., 5].astype(np.int32)
    return blines, cls, confs


class TextDetector:
    lang_list = ["eng", "ja", "unknown"]
    langcls2idx = {"eng": 0, "ja": 1, "unknown": 2}

    def __init__(
        self,
        model_path,
        input_size=1024,
        device="cpu",
        half=False,
        nms_thresh=0.35,
        conf_thresh=0.4,
        mask_thresh=0.3,
        act="leaky",
    ):
        super(TextDetector, self).__init__()
        cuda = device == "cuda"

        if Path(model_path).suffix == ".onnx":
            self.model = cv2.dnn.readNetFromONNX(model_path)
            self.net = TextDetBaseDNN(input_size, model_path)
            self.backend = "opencv"
        else:
            self.net = TextDetBase(model_path, device=device, act=act)
            self.backend = "torch"

        if isinstance(input_size, int):
            input_size = (input_size, input_size)
        self.input_size = input_size
        self.device = device
        self.half = half
        self.conf_thresh = conf_thresh
        self.nms_thresh = nms_thresh
        self.seg_rep = SegDetectorRepresenter(thresh=0.3)

    @torch.no_grad()
    def __call__(self, img, refine_mode=REFINEMASK_INPAINT, keep_undetected_mask=False):
        img_in, ratio, dw, dh = preprocess_img(
            img,
            input_size=self.input_size,
            device=self.device,
            half=self.half,
            to_tensor=self.backend == "torch",
        )
        im_h, im_w = img.shape[:2]

        blks, mask, lines_map = self.net(img_in)

        resize_ratio = (im_w / (self.input_size[0] - dw), im_h / (self.input_size[1] - dh))
        blks = postprocess_yolo(blks, self.conf_thresh, self.nms_thresh, resize_ratio)

        if self.backend == "opencv":
            if mask.shape[1] == 2:  # some version of opencv spit out reversed result
                tmp = mask
                mask = lines_map
                lines_map = tmp
        mask = postprocess_mask(mask)

        lines, scores = self.seg_rep(self.input_size, lines_map)
        box_thresh = 0.6
        idx = np.where(scores[0] > box_thresh)
        lines, scores = lines[0][idx], scores[0][idx]

        # map output to input img
        mask = mask[: mask.shape[0] - dh, : mask.shape[1] - dw]
        mask = cv2.resize(mask, (im_w, im_h), interpolation=cv2.INTER_LINEAR)
        if lines.size == 0:
            lines = []
        else:
            lines = lines.astype(np.float64)
            lines[..., 0] *= resize_ratio[0]
            lines[..., 1] *= resize_ratio[1]
            lines = lines.astype(np.int32)
        blk_list = group_output(blks, lines, im_w, im_h, mask)
        mask_refined = refine_mask(img, mask, blk_list, refine_mode=refine_mode)
        if keep_undetected_mask:
            mask_refined = refine_undetected_mask(
                img, mask, mask_refined, blk_list, refine_mode=refine_mode
            )

        return mask, mask_refined, blk_list


def traverse_by_dict(img_dir_list, dict_dir):
    if isinstance(img_dir_list, str):
        img_dir_list = [img_dir_list]
    imglist = []
    for img_dir in img_dir_list:
        imglist += find_all_imgs(img_dir, abs_path=True)
    for img_path in tqdm(imglist):
        imgname = osp.basename(img_path)
        imname = imgname.replace(Path(imgname).suffix, "")
        mask_path = osp.join(dict_dir, "mask-" + imname + ".png")
        with open(osp.join(dict_dir, imname + ".json"), "r", encoding="utf8") as f:
            blk_dict_list = json.loads(f.read())
            blk_list = [TextBlock(**blk_dict) for blk_dict in blk_dict_list]
        img = cv2.imread(img_path)
        mask = cv2.imread(mask_path, cv2.IMREAD_GRAYSCALE)
        mask = refine_mask(img, mask, blk_list)

        visualize_textblocks(img, blk_list)
        cv2.imshow("im", img)
        cv2.imshow("mask", mask)
        cv2.waitKey(0)



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Common modules
"""

import math
import warnings

import torch
import torch.nn as nn


def autopad(k, p=None):  # kernel, padding
    # Pad to 'same'
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


class Conv(nn.Module):
    # Standard convolution
    def __init__(
        self, c1, c2, k=1, s=1, p=None, g=1, act=True
    ):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        if isinstance(act, bool):
            self.act = (
                nn.SiLU() if act is True else (act if isinstance(act, nn.Module) else nn.Identity())
            )
        elif isinstance(act, str):
            if act == "leaky":
                self.act = nn.LeakyReLU(0.1, inplace=True)
            elif act == "relu":
                self.act = nn.ReLU(inplace=True)
            else:
                self.act = None

    def forward(self, x):
        return self.act(self.bn(self.conv(x)))

    def forward_fuse(self, x):
        return self.act(self.conv(x))


class DWConv(Conv):
    # Depth-wise convolution class
    def __init__(
        self, c1, c2, k=1, s=1, act=True
    ):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__(c1, c2, k, s, g=math.gcd(c1, c2), act=act)


class TransformerLayer(nn.Module):
    # Transformer layer https://arxiv.org/abs/2010.11929 (LayerNorm layers removed for better performance)
    def __init__(self, c, num_heads):
        super().__init__()
        self.q = nn.Linear(c, c, bias=False)
        self.k = nn.Linear(c, c, bias=False)
        self.v = nn.Linear(c, c, bias=False)
        self.ma = nn.MultiheadAttention(embed_dim=c, num_heads=num_heads)
        self.fc1 = nn.Linear(c, c, bias=False)
        self.fc2 = nn.Linear(c, c, bias=False)

    def forward(self, x):
        x = self.ma(self.q(x), self.k(x), self.v(x))[0] + x
        x = self.fc2(self.fc1(x)) + x
        return x


class TransformerBlock(nn.Module):
    # Vision Transformer https://arxiv.org/abs/2010.11929
    def __init__(self, c1, c2, num_heads, num_layers):
        super().__init__()
        self.conv = None
        if c1 != c2:
            self.conv = Conv(c1, c2)
        self.linear = nn.Linear(c2, c2)  # learnable position embedding
        self.tr = nn.Sequential(*(TransformerLayer(c2, num_heads) for _ in range(num_layers)))
        self.c2 = c2

    def forward(self, x):
        if self.conv is not None:
            x = self.conv(x)
        b, _, w, h = x.shape
        p = x.flatten(2).permute(2, 0, 1)
        return self.tr(p + self.linear(p)).permute(1, 2, 0).reshape(b, self.c2, w, h)


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(
        self, c1, c2, shortcut=True, g=1, e=0.5, act=True
    ):  # ch_in, ch_out, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1, act=act)
        self.cv2 = Conv(c_, c2, 3, 1, g=g, act=act)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    # CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    def __init__(
        self, c1, c2, n=1, shortcut=True, g=1, e=0.5
    ):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = nn.Conv2d(c1, c_, 1, 1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, 1, 1, bias=False)
        self.cv4 = Conv(2 * c_, c2, 1, 1)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.SiLU()
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)))

    def forward(self, x):
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))


class C3(nn.Module):
    # CSP Bottleneck with 3 convolutions
    def __init__(
        self, c1, c2, n=1, shortcut=True, g=1, e=0.5, act=True
    ):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1, act=act)
        self.cv2 = Conv(c1, c_, 1, 1, act=act)
        self.cv3 = Conv(2 * c_, c2, 1, act=act)  # act=FReLU(c2)
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, e=1.0, act=act) for _ in range(n)))
        # self.m = nn.Sequential(*[CrossConv(c_, c_, 3, 1, g, 1.0, shortcut) for _ in range(n)])

    def forward(self, x):
        return self.cv3(torch.cat((self.m(self.cv1(x)), self.cv2(x)), dim=1))


class C3TR(C3):
    # C3 module with TransformerBlock()
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)
        self.m = TransformerBlock(c_, c_, 4, n)


class C3SPP(C3):
    # C3 module with SPP()
    def __init__(self, c1, c2, k=(5, 9, 13), n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)
        self.m = SPP(c_, c_, k)


class C3Ghost(C3):
    # C3 module with GhostBottleneck()
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)  # hidden channels
        self.m = nn.Sequential(*(GhostBottleneck(c_, c_) for _ in range(n)))


class SPP(nn.Module):
    # Spatial Pyramid Pooling (SPP) layer https://arxiv.org/abs/1406.4729
    def __init__(self, c1, c2, k=(5, 9, 13)):
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * (len(k) + 1), c2, 1, 1)
        self.m = nn.ModuleList([nn.MaxPool2d(kernel_size=x, stride=1, padding=x // 2) for x in k])

    def forward(self, x):
        x = self.cv1(x)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # suppress torch 1.9.0 max_pool2d() warning
            return self.cv2(torch.cat([x] + [m(x) for m in self.m], 1))


class SPPF(nn.Module):
    # Spatial Pyramid Pooling - Fast (SPPF) layer for YOLOv5 by Glenn Jocher
    def __init__(self, c1, c2, k=5):  # equivalent to SPP(k=(5, 9, 13))
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * 4, c2, 1, 1)
        self.m = nn.MaxPool2d(kernel_size=k, stride=1, padding=k // 2)

    def forward(self, x):
        x = self.cv1(x)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # suppress torch 1.9.0 max_pool2d() warning
            y1 = self.m(x)
            y2 = self.m(y1)
            return self.cv2(torch.cat([x, y1, y2, self.m(y2)], 1))


class Focus(nn.Module):
    # Focus wh information into c-space
    def __init__(
        self, c1, c2, k=1, s=1, p=None, g=1, act=True
    ):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = Conv(c1 * 4, c2, k, s, p, g, act)
        # self.contract = Contract(gain=2)

    def forward(self, x):  # x(b,c,w,h) -> y(b,4c,w/2,h/2)
        return self.conv(
            torch.cat(
                [x[..., ::2, ::2], x[..., 1::2, ::2], x[..., ::2, 1::2], x[..., 1::2, 1::2]], 1
            )
        )
        # return self.conv(self.contract(x))


class GhostConv(nn.Module):
    # Ghost Convolution https://github.com/huawei-noah/ghostnet
    def __init__(self, c1, c2, k=1, s=1, g=1, act=True):  # ch_in, ch_out, kernel, stride, groups
        super().__init__()
        c_ = c2 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, k, s, None, g, act)
        self.cv2 = Conv(c_, c_, 5, 1, None, c_, act)

    def forward(self, x):
        y = self.cv1(x)
        return torch.cat([y, self.cv2(y)], 1)


class GhostBottleneck(nn.Module):
    # Ghost Bottleneck https://github.com/huawei-noah/ghostnet
    def __init__(self, c1, c2, k=3, s=1):  # ch_in, ch_out, kernel, stride
        super().__init__()
        c_ = c2 // 2
        self.conv = nn.Sequential(
            GhostConv(c1, c_, 1, 1),  # pw
            DWConv(c_, c_, k, s, act=False) if s == 2 else nn.Identity(),  # dw
            GhostConv(c_, c2, 1, 1, act=False),
        )  # pw-linear
        self.shortcut = (
            nn.Sequential(DWConv(c1, c1, k, s, act=False), Conv(c1, c2, 1, 1, act=False))
            if s == 2
            else nn.Identity()
        )

    def forward(self, x):
        return self.conv(x) + self.shortcut(x)


class Contract(nn.Module):
    # Contract width-height into channels, i.e. x(1,64,80,80) to x(1,256,40,40)
    def __init__(self, gain=2):
        super().__init__()
        self.gain = gain

    def forward(self, x):
        b, c, h, w = x.size()  # assert (h / s == 0) and (W / s == 0), 'Indivisible gain'
        s = self.gain
        x = x.view(b, c, h // s, s, w // s, s)  # x(1,64,40,2,40,2)
        x = x.permute(0, 3, 5, 1, 2, 4).contiguous()  # x(1,2,2,64,40,40)
        return x.view(b, c * s * s, h // s, w // s)  # x(1,256,40,40)


class Expand(nn.Module):
    # Expand channels into width-height, i.e. x(1,64,80,80) to x(1,16,160,160)
    def __init__(self, gain=2):
        super().__init__()
        self.gain = gain

    def forward(self, x):
        b, c, h, w = x.size()  # assert C / s ** 2 == 0, 'Indivisible gain'
        s = self.gain
        x = x.view(b, s, s, c // s**2, h, w)  # x(1,2,2,16,80,80)
        x = x.permute(0, 3, 4, 1, 5, 2).contiguous()  # x(1,16,80,2,80,2)
        return x.view(b, c // s**2, h * s, w * s)  # x(1,16,160,160)


class Concat(nn.Module):
    # Concatenate a list of tensors along dimension
    def __init__(self, dimension=1):
        super().__init__()
        self.d = dimension

    def forward(self, x):
        return torch.cat(x, self.d)


class Classify(nn.Module):
    # Classification head, i.e. x(b,c1,20,20) to x(b,c2)
    def __init__(
        self, c1, c2, k=1, s=1, p=None, g=1
    ):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.aap = nn.AdaptiveAvgPool2d(1)  # to x(b,c1,1,1)
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g)  # to x(b,c2,1,1)
        self.flat = nn.Flatten()

    def forward(self, x):
        z = torch.cat([self.aap(y) for y in (x if isinstance(x, list) else [x])], 1)  # cat if list
        return self.flat(self.conv(z))  # flatten to x(b,c2)



import os
from pathlib import Path
from hashlib import sha256

import requests
import tqdm
import torch


MODEL_URL = "https://github.com/zyddnys/manga-image-translator/releases/download/beta-0.3/"
TORCH_MODEL_NAME = "comictextdetector.pt"
TORCH_SHA256 = "1f90fa60aeeb1eb82e2ac1167a66bf139a8a61b8780acd351ead55268540cccb"
CV2_MODEL_NAME = "comictextdetector.pt.onnx"
CV2_SHA256 = "1a86ace74961413cbd650002e7bb4dcec4980ffa21b2f19b86933372071d718f"


def check_hash(file_path: Path, sha_hash: str) -> bool:
    """
    Check if the file has the expected sha256 hash.

    :param file_path: The path to the file.
    :param sha_hash: The expected sha256 hash.
    :return: True if the file has the expected hash.
    """
    return sha256(file_path.read_bytes()).hexdigest() == sha_hash


def download_file(url: str, save_dir: Path, sha_hash: str = None) -> Path | None:
    """
    Download a file from a url and return the path to the downloaded file.

    :param url: The url to download the file from.
    :param save_dir: The directory where the file will be saved.
    :param sha_hash: The sha256 hash of the file. If it doesn't match, the download will be aborted.
    :return: The path to the downloaded file.
    """
    response = requests.get(url, stream=True)
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        print(f"Error downloading file from url: {url}\n{e}")
        print("Manually download the file and save the path to it in a profile's settings.")
        return None

    file_name = os.path.basename(url)
    save_path = save_dir / file_name

    # Ensure the file is not already downloaded, and the folder exists.
    save_dir.mkdir(parents=True, exist_ok=True)
    if save_path.is_file():
        save_path.unlink()
    elif save_path.is_dir():
        save_path.rmdir()

    file_size = int(response.headers.get("Content-Length", 0))
    chunk_size = 8192  # 8KB

    with open(save_path, "wb") as f, tqdm.tqdm(total=file_size, unit="B", unit_scale=True) as pbar:
        for chunk in response.iter_content(chunk_size=chunk_size):
            if chunk:
                f.write(chunk)
                pbar.update(len(chunk))

    if not save_path.exists():
        print(f"Error downloading file from url: {url}\nFailed to save the file to {save_path}")
    if sha_hash is not None:
        if not check_hash(save_path, sha_hash):
            print(
                f"Error downloading file from url: {url}\nThe file content is different from expected."
            )
            save_path.unlink()
            return None

    return save_path


def download_torch_model(cache_dir: Path) -> Path | None:
    """
    Download the torch model and return the path to the downloaded file.

    :param cache_dir: The directory where the file will be saved.
    :return: The path to the downloaded file. Or None if the download failed.
    """
    print("Downloading Torch model for CUDA...")
    return download_file(MODEL_URL + TORCH_MODEL_NAME, cache_dir, sha_hash=TORCH_SHA256)


def download_cv2_model(cache_dir: Path) -> Path | None:
    """
    Download the cv2 model and return the path to the downloaded file.

    :param cache_dir: The directory where the file will be saved.
    :return: The path to the downloaded file. Or None if the download failed.
    """
    print("Downloading OpenCV model for CPU...")
    return download_file(MODEL_URL + CV2_MODEL_NAME, cache_dir, sha_hash=CV2_SHA256)


def download_models(config, force: bool, cuda: bool, cpu: bool):
    """
    Download the models and save the paths to them in the config.
    If neither cuda nor cpu is set, use cuda if available.

    :param config: The config to save the paths in.
    :param force: If True, overwrite the paths in the config.
    :param cuda: If True, only download the cuda model.
    :param cpu: If True, only download the cpu model.
    :return:
    """
    if not cuda and not cpu:
        cuda = torch.cuda.is_available()
        cpu = not cuda

    cache_dir = config.get_model_cache_dir()
    # Check if they aren't set already.
    if cuda:
        if force or config.default_torch_model_path is None:
            config.default_torch_model_path = download_torch_model(cache_dir)
        else:
            print(
                f"Skipping Torch model download, path already set to: {config.default_torch_model_path}"
            )
    if cpu:
        if force or config.default_cv2_model_path is None:
            config.default_cv2_model_path = download_cv2_model(cache_dir)
        else:
            print(
                f"Skipping OpenCV model download, path already set to: {config.default_cv2_model_path}"
            )

    config.save()



# import onnx
# import onnxsim
# import torch
# import torch.nn as nn
# from models.yolov5.common import Conv
# from models.yolov5.yolo import Detect
#
#
# class SiLU(nn.Module):  # export-friendly version of nn.SiLU()
#     @staticmethod
#     def forward(x):
#         return x * torch.sigmoid(x)
#
#
# def concate_models(blk_weights, seg_weights, det_weights, save_path):
#     textdetector_dict = dict()
#     textdetector_dict["blk_det"] = torch.load(blk_weights, map_location="cpu")
#     textdetector_dict["text_seg"] = torch.load(seg_weights, map_location="cpu")["weights"]
#     textdetector_dict["text_det"] = torch.load(det_weights, map_location="cpu")["weights"]
#     torch.save(textdetector_dict, save_path)
#
#
# def export_onnx(model, im, file, opset, train=False, simplify=True, dynamic=False, inplace=False):
#     # YOLOv5 ONNX export
#     f = file + ".onnx"
#     for k, m in model.named_modules():
#         if isinstance(m, Conv):  # assign export-friendly activations
#             if isinstance(m.act, nn.SiLU):
#                 m.act = SiLU()
#         elif isinstance(m, Detect):
#             m.inplace = inplace
#             m.onnx_dynamic = False
#     torch.onnx.export(
#         model,
#         im,
#         f,
#         verbose=False,
#         opset_version=opset,
#         training=torch.onnx.TrainingMode.TRAINING if train else torch.onnx.TrainingMode.EVAL,
#         do_constant_folding=not train,
#         input_names=["images"],
#         output_names=["blk", "seg", "det"],
#         dynamic_axes={
#             "images": {0: "batch", 2: "height", 3: "width"},  # shape(1,3,640,640)
#             "output": {0: "batch", 1: "anchors"},  # shape(1,25200,85)
#         }
#         if dynamic
#         else None,
#     )
#
#     # Checks
#     model_onnx = onnx.load(f)  # load onnx model
#     onnx.checker.check_model(model_onnx)  # check onnx model
#
#     model_onnx, check = onnxsim.simplify(
#         model_onnx, dynamic_input_shape=dynamic, input_shapes={"images": list(im.shape)} if dynamic else None
#     )
#     assert check, "assert check failed"
#     onnx.save(model_onnx, f)



import copy
import math
from typing import List

import cv2
import numpy as np
from shapely.geometry import Polygon
from .imgproc_utils import union_area, xywh2xyxypoly, rotate_polygons

LANG_LIST = ["eng", "ja", "unknown"]
LANGCLS2IDX = {"eng": 0, "ja": 1, "unknown": 2}


class TextBlock(object):
    def __init__(
        self,
        xyxy: List,
        lines: List = None,
        language: str = "unknown",
        vertical: bool = False,
        font_size: float = -1,
        distance: List = None,
        angle: int = 0,
        vec: List = None,
        norm: float = -1,
        merged: bool = False,
        weight: float = -1,
        text: List = None,
        translation: str = "",
        fg_r=0,
        fg_g=0,
        fg_b=0,
        bg_r=0,
        bg_g=0,
        bg_b=0,
        line_spacing=1.0,
        font_family: str = "",
        bold: bool = False,
        underline: bool = False,
        italic: bool = False,
        alignment: int = -1,
        alpha: float = 255,
        rich_text: str = "",
        _bounding_rect: List = None,
        accumulate_color=True,
        default_stroke_width=0.2,
        target_lang: str = "",
        **kwargs
    ) -> None:
        self.xyxy = [int(num) for num in xyxy]  # boundingbox of textblock
        self.lines = [] if lines is None else lines  # polygons of textlines
        self.vertical = vertical  # orientation of textlines
        self.language = language
        self.font_size = font_size  # font pixel size
        self.distance = (
            None if distance is None else np.array(distance, np.float64)
        )  # distance between textlines and "origin"
        self.angle = angle  # rotation angle of textlines

        self.vec = None if vec is None else np.array(vec, np.float64)  # primary vector of textblock
        self.norm = norm  # primary norm of textblock
        self.merged = merged
        self.weight = weight

        self.text = text if text is not None else []
        self.prob = 1

        self.translation = translation

        # note they're accumulative rgb values of textlines
        self.fg_r = fg_r
        self.fg_g = fg_g
        self.fg_b = fg_b
        self.bg_r = bg_r
        self.bg_g = bg_g
        self.bg_b = bg_b

        # self.stroke_width = stroke_width
        self.font_family: str = font_family
        self.bold: bool = bold
        self.underline: bool = underline
        self.italic: bool = italic
        self.alpha = alpha
        self.rich_text = rich_text
        self.line_spacing = line_spacing
        # self.alignment = alignment
        self._alignment = alignment
        self._target_lang = target_lang

        self._bounding_rect = _bounding_rect
        self.default_stroke_width = default_stroke_width
        self.accumulate_color = accumulate_color

    def adjust_bbox(self, with_bbox=False):
        lines = self.lines_array().astype(np.int32)
        if with_bbox:
            self.xyxy[0] = min(lines[..., 0].min(), self.xyxy[0])
            self.xyxy[1] = min(lines[..., 1].min(), self.xyxy[1])
            self.xyxy[2] = max(lines[..., 0].max(), self.xyxy[2])
            self.xyxy[3] = max(lines[..., 1].max(), self.xyxy[3])
        else:
            self.xyxy[0] = lines[..., 0].min()
            self.xyxy[1] = lines[..., 1].min()
            self.xyxy[2] = lines[..., 0].max()
            self.xyxy[3] = lines[..., 1].max()

    def sort_lines(self):
        if self.distance is not None:
            idx = np.argsort(self.distance)
            self.distance = self.distance[idx]
            lines = np.array(self.lines, dtype=np.int32)
            self.lines = lines[idx].tolist()

    def lines_array(self, dtype=np.float64):
        return np.array(self.lines, dtype=dtype)

    def aspect_ratio(self) -> float:
        min_rect = self.min_rect()
        middle_pnts = (min_rect[:, [1, 2, 3, 0]] + min_rect) / 2
        norm_v = np.linalg.norm(middle_pnts[:, 2] - middle_pnts[:, 0])
        norm_h = np.linalg.norm(middle_pnts[:, 1] - middle_pnts[:, 3])
        return norm_v / norm_h

    def center(self):
        xyxy = np.array(self.xyxy)
        return (xyxy[:2] + xyxy[2:]) / 2

    def min_rect(self, rotate_back=True):
        angled = self.angle != 0
        center = self.center()
        polygons = self.lines_array().reshape(-1, 8)
        if angled:
            polygons = rotate_polygons(center, polygons, self.angle)
        min_x = polygons[:, ::2].min()
        min_y = polygons[:, 1::2].min()
        max_x = polygons[:, ::2].max()
        max_y = polygons[:, 1::2].max()
        min_bbox = np.array([[min_x, min_y, max_x, min_y, max_x, max_y, min_x, max_y]])
        if angled and rotate_back:
            min_bbox = rotate_polygons(center, min_bbox, -self.angle)
        return min_bbox.reshape(-1, 4, 2).astype(np.int64)

    # equivalent to qt's boundingRect, ignore angle
    def bounding_rect(self):
        if self._bounding_rect is None:
            # if True:
            min_bbox = self.min_rect(rotate_back=False)[0]
            x, y = min_bbox[0]
            w, h = min_bbox[2] - min_bbox[0]
            return [x, y, w, h]
        return self._bounding_rect

    def __getattribute__(self, name: str):
        if name == "pts":
            return self.lines_array()
        # else:
        return object.__getattribute__(self, name)

    def __len__(self):
        return len(self.lines)

    def __getitem__(self, idx):
        return self.lines[idx]

    def to_dict(self):
        blk_dict = copy.deepcopy(vars(self))
        return blk_dict

    def get_transformed_region(self, img, idx, textheight) -> np.ndarray:
        im_h, im_w = img.shape[:2]
        direction = "v" if self.vertical else "h"
        src_pts = np.array(self.lines[idx], dtype=np.float64)

        if self.language == "eng" or (self.language == "unknown" and not self.vertical):
            e_size = self.font_size / 3
            src_pts[..., 0] += np.array([-e_size, e_size, e_size, -e_size])
            src_pts[..., 1] += np.array([-e_size, -e_size, e_size, e_size])
            src_pts[..., 0] = np.clip(src_pts[..., 0], 0, im_w)
            src_pts[..., 1] = np.clip(src_pts[..., 1], 0, im_h)

        middle_pnt = (src_pts[[1, 2, 3, 0]] + src_pts) / 2
        vec_v = middle_pnt[2] - middle_pnt[0]  # vertical vectors of textlines
        vec_h = middle_pnt[1] - middle_pnt[3]  # horizontal vectors of textlines
        ratio = np.linalg.norm(vec_v) / np.linalg.norm(vec_h)

        if direction == "h":
            h = int(textheight)
            w = int(round(textheight / ratio))
            dst_pts = np.array([[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1]]).astype(np.float32)
            M, _ = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0)
            region = cv2.warpPerspective(img, M, (w, h))
        elif direction == "v":
            w = int(textheight)
            h = int(round(textheight * ratio))
            dst_pts = np.array([[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1]]).astype(np.float32)
            M, _ = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0)
            region = cv2.warpPerspective(img, M, (w, h))
            region = cv2.rotate(region, cv2.ROTATE_90_COUNTERCLOCKWISE)
        # cv2.imshow('region'+str(idx), region)
        # cv2.waitKey(0)
        return region

    def get_text(self):
        if isinstance(self.text, str):
            return self.text
        return " ".join(self.text).strip()

    def set_font_colors(self, frgb, srgb, accumulate=True):
        self.accumulate_color = accumulate
        num_lines = len(self.lines) if accumulate and len(self.lines) > 0 else 1
        # set font color
        frgb = np.array(frgb) * num_lines
        self.fg_r, self.fg_g, self.fg_b = frgb
        # set stroke color
        srgb = np.array(srgb) * num_lines
        self.bg_r, self.bg_g, self.bg_b = srgb

    def get_font_colors(self, bgr=False):
        num_lines = len(self.lines)
        frgb = np.array([self.fg_r, self.fg_g, self.fg_b])
        brgb = np.array([self.bg_r, self.bg_g, self.bg_b])
        if self.accumulate_color:
            if num_lines > 0:
                frgb = (frgb / num_lines).astype(np.int32)
                brgb = (brgb / num_lines).astype(np.int32)
                if bgr:
                    return frgb[::-1], brgb[::-1]
                else:
                    return frgb, brgb
            else:
                return [0, 0, 0], [0, 0, 0]
        else:
            return frgb, brgb

    def xywh(self):
        x, y, w, h = self.xyxy
        return [x, y, w - x, h - y]

    # alignleft: 0, center: 1, right: 2
    def alignment(self):
        if self._alignment >= 0:
            return self._alignment
        elif self.vertical:
            return 0
        lines = self.lines_array()
        if len(lines) == 1:
            return 0
        angled = self.angle != 0
        polygons = lines.reshape(-1, 8)
        if angled:
            polygons = rotate_polygons((0, 0), polygons, self.angle)
        polygons = polygons.reshape(-1, 4, 2)

        left_std = np.std(polygons[:, 0, 0])
        # right_std = np.std(polygons[:, 1, 0])
        center_std = np.std((polygons[:, 0, 0] + polygons[:, 1, 0]) / 2)
        if left_std < center_std:
            return 0
        else:
            return 1

    def target_lang(self):
        return self.target_lang

    @property
    def stroke_width(self):
        var = np.array([self.fg_r, self.fg_g, self.fg_b]) - np.array(
            [self.bg_r, self.bg_g, self.bg_b]
        )
        var = np.abs(var).sum()
        if var > 40:
            return self.default_stroke_width
        return 0


def sort_textblk_list(blk_list: List[TextBlock], im_w: int, im_h: int) -> List[TextBlock]:
    if len(blk_list) == 0:
        return blk_list
    num_ja = 0
    xyxy = []
    for blk in blk_list:
        if blk.language == "ja":
            num_ja += 1
        xyxy.append(blk.xyxy)
    xyxy = np.array(xyxy)
    flip_lr = num_ja > len(blk_list) / 2
    im_oriw = im_w
    if im_w > im_h:
        im_w /= 2
    num_gridy, num_gridx = 4, 3
    img_area = im_h * im_w
    center_x = (xyxy[:, 0] + xyxy[:, 2]) / 2
    if flip_lr:
        if im_w != im_oriw:
            center_x = im_oriw - center_x
        else:
            center_x = im_w - center_x
    grid_x = (center_x / im_w * num_gridx).astype(np.int32)
    center_y = (xyxy[:, 1] + xyxy[:, 3]) / 2
    grid_y = (center_y / im_h * num_gridy).astype(np.int32)
    grid_indices = grid_y * num_gridx + grid_x
    grid_weights = (
        grid_indices * img_area
        + 1.2 * (center_x - grid_x * im_w / num_gridx)
        + (center_y - grid_y * im_h / num_gridy)
    )
    if im_w != im_oriw:
        grid_weights[np.where(grid_x >= num_gridx)] += img_area * num_gridy * num_gridx

    for blk, weight in zip(blk_list, grid_weights):
        blk.weight = weight
    blk_list.sort(key=lambda blk: blk.weight)
    return blk_list


def examine_textblk(blk: TextBlock, im_w: int, im_h: int, sort: bool = False) -> None:
    lines = blk.lines_array()
    middle_pnts = (lines[:, [1, 2, 3, 0]] + lines) / 2
    vec_v = middle_pnts[:, 2] - middle_pnts[:, 0]  # vertical vectors of textlines
    vec_h = middle_pnts[:, 1] - middle_pnts[:, 3]  # horizontal vectors of textlines
    # if sum of vertical vectors is longer, then text orientation is vertical, and vice versa.
    center_pnts = (lines[:, 0] + lines[:, 2]) / 2
    v = np.sum(vec_v, axis=0)
    h = np.sum(vec_h, axis=0)
    norm_v, norm_h = np.linalg.norm(v), np.linalg.norm(h)
    if blk.language == "ja":
        vertical = norm_v > norm_h
    else:
        vertical = norm_v > norm_h * 2
    # calcuate distance between textlines and origin
    if vertical:
        primary_vec, primary_norm = v, norm_v
        distance_vectors = center_pnts - np.array(
            [[im_w, 0]], dtype=np.float64
        )  # vertical manga text is read from right to left, so origin is (imw, 0)
        font_size = int(round(norm_h / len(lines)))
    else:
        primary_vec, primary_norm = h, norm_h
        distance_vectors = center_pnts - np.array([[0, 0]], dtype=np.float64)
        font_size = int(round(norm_v / len(lines)))

    rotation_angle = int(
        math.atan2(primary_vec[1], primary_vec[0]) / math.pi * 180
    )  # rotation angle of textlines
    distance = np.linalg.norm(
        distance_vectors, axis=1
    )  # distance between textlinecenters and origin
    rad_matrix = np.arccos(
        np.einsum("ij, j->i", distance_vectors, primary_vec) / (distance * primary_norm)
    )
    distance = np.abs(np.sin(rad_matrix) * distance)
    blk.lines = lines.astype(np.int32).tolist()
    blk.distance = distance
    blk.angle = rotation_angle
    if vertical:
        blk.angle -= 90
    if abs(blk.angle) < 3:
        blk.angle = 0
    blk.font_size = font_size
    blk.vertical = vertical
    blk.vec = primary_vec
    blk.norm = primary_norm
    if sort:
        blk.sort_lines()


def try_merge_textline(blk: TextBlock, blk2: TextBlock, fntsize_tol=1.3, distance_tol=2) -> bool:
    if blk2.merged:
        return False
    fntsize_div = blk.font_size / blk2.font_size
    num_l1, num_l2 = len(blk), len(blk2)
    fntsz_avg = (blk.font_size * num_l1 + blk2.font_size * num_l2) / (num_l1 + num_l2)
    vec_prod = blk.vec @ blk2.vec
    vec_sum = blk.vec + blk2.vec
    cos_vec = vec_prod / blk.norm / blk2.norm
    distance = blk2.distance[-1] - blk.distance[-1]
    distance_p1 = np.linalg.norm(np.array(blk2.lines[-1][0]) - np.array(blk.lines[-1][0]))
    l1, l2 = Polygon(blk.lines[-1]), Polygon(blk2.lines[-1])
    if not l1.intersects(l2):
        if fntsize_div > fntsize_tol or 1 / fntsize_div > fntsize_tol:
            return False
        if abs(cos_vec) < 0.866:  # cos30
            return False
        if distance > distance_tol * fntsz_avg or distance_p1 > fntsz_avg * 2.5:
            return False
    # merge
    blk.lines.append(blk2.lines[0])
    blk.vec = vec_sum
    blk.angle = int(round(np.rad2deg(math.atan2(vec_sum[1], vec_sum[0]))))
    if blk.vertical:
        blk.angle -= 90
    blk.norm = np.linalg.norm(vec_sum)
    blk.distance = np.append(blk.distance, blk2.distance[-1])
    blk.font_size = fntsz_avg
    blk2.merged = True
    return True


def merge_textlines(blk_list: List[TextBlock]) -> List[TextBlock]:
    if len(blk_list) < 2:
        return blk_list
    blk_list.sort(key=lambda blk: blk.distance[0])
    merged_list = []
    for ii, current_blk in enumerate(blk_list):
        if current_blk.merged:
            continue
        for jj, blk in enumerate(blk_list[ii + 1 :]):
            try_merge_textline(current_blk, blk)
        merged_list.append(current_blk)
    for blk in merged_list:
        blk.adjust_bbox(with_bbox=False)
    return merged_list


def split_textblk(blk: TextBlock):
    font_size, distance, lines = blk.font_size, blk.distance, blk.lines
    l0 = np.array(blk.lines[0])
    lines.sort(key=lambda line: np.linalg.norm(np.array(line[0]) - l0[0]))
    distance_tol = font_size * 2
    current_blk = copy.deepcopy(blk)
    current_blk.lines = [l0]
    sub_blk_list = [current_blk]
    textblock_splitted = False
    for jj, line in enumerate(lines[1:]):
        l1, l2 = Polygon(lines[jj]), Polygon(line)
        split = False
        if not l1.intersects(l2):
            line_disance = abs(distance[jj + 1] - distance[jj])
            if line_disance > distance_tol:
                split = True
            elif blk.vertical and abs(blk.angle) < 15:
                if len(current_blk.lines) > 1 or line_disance > font_size:
                    split = abs(lines[jj][0][1] - line[0][1]) > font_size
        if split:
            current_blk = copy.deepcopy(current_blk)
            current_blk.lines = [line]
            sub_blk_list.append(current_blk)
        else:
            current_blk.lines.append(line)
    if len(sub_blk_list) > 1:
        textblock_splitted = True
        for current_blk in sub_blk_list:
            current_blk.adjust_bbox(with_bbox=False)
    return textblock_splitted, sub_blk_list


def group_output(blks, lines, im_w, im_h, mask=None, sort_blklist=True) -> List[TextBlock]:
    blk_list: List[TextBlock] = []
    scattered_lines = {"ver": [], "hor": []}
    for bbox, cls, conf in zip(*blks):
        # cls could give wrong result
        blk_list.append(TextBlock(bbox, language=LANG_LIST[cls]))

    # step1: filter & assign lines to textblocks
    bbox_score_thresh = 0.4
    mask_score_thresh = 0.1
    for ii, line in enumerate(lines):
        bx1, bx2 = line[:, 0].min(), line[:, 0].max()
        by1, by2 = line[:, 1].min(), line[:, 1].max()
        bbox_score, bbox_idx = -1, -1
        line_area = (by2 - by1) * (bx2 - bx1)
        for jj, blk in enumerate(blk_list):
            score = union_area(blk.xyxy, [bx1, by1, bx2, by2]) / line_area
            if bbox_score < score:
                bbox_score = score
                bbox_idx = jj
        if bbox_score > bbox_score_thresh:
            blk_list[bbox_idx].lines.append(line)
        else:  # if no textblock was assigned, check whether there is "enough" textmask
            if mask is not None:
                mask_score = mask[by1:by2, bx1:bx2].mean() / 255
                if mask_score < mask_score_thresh:
                    continue
            blk = TextBlock([bx1, by1, bx2, by2], [line])
            examine_textblk(blk, im_w, im_h, sort=False)
            if blk.vertical:
                scattered_lines["ver"].append(blk)
            else:
                scattered_lines["hor"].append(blk)

    # step2: filter textblocks, sort & split textlines
    final_blk_list = []
    for blk in blk_list:
        # filter textblocks
        if len(blk.lines) == 0:
            bx1, by1, bx2, by2 = blk.xyxy
            if mask is not None:
                mask_score = mask[by1:by2, bx1:bx2].mean() / 255
                if mask_score < mask_score_thresh:
                    continue
            xywh = np.array([[bx1, by1, bx2 - bx1, by2 - by1]])
            blk.lines = xywh2xyxypoly(xywh).reshape(-1, 4, 2).tolist()
        examine_textblk(blk, im_w, im_h, sort=True)

        # split manga text if there is a distance gap
        textblock_splitted = False
        if len(blk.lines) > 1:
            if blk.language == "ja":
                textblock_splitted = True
            elif blk.vertical:
                textblock_splitted = True
        if textblock_splitted:
            textblock_splitted, sub_blk_list = split_textblk(blk)
        else:
            sub_blk_list = [blk]
        # modify textblock to fit its textlines
        if not textblock_splitted:
            for blk in sub_blk_list:
                blk.adjust_bbox(with_bbox=True)
        final_blk_list += sub_blk_list

    # step3: merge scattered lines, sort textblocks by "grid"
    final_blk_list += merge_textlines(scattered_lines["hor"])
    final_blk_list += merge_textlines(scattered_lines["ver"])
    if sort_blklist:
        final_blk_list = sort_textblk_list(final_blk_list, im_w, im_h)

    for blk in final_blk_list:
        if blk.language == "eng" and not blk.vertical:
            num_lines = len(blk.lines)
            if num_lines == 0:
                continue
            # blk.line_spacing = blk.bounding_rect()[3] / num_lines / blk.font_size
            expand_size = max(int(blk.font_size * 0.1), 2)
            rad = np.deg2rad(blk.angle)
            shifted_vec = np.array([[[-1, -1], [1, -1], [1, 1], [-1, 1]]])
            shifted_vec = shifted_vec * np.array([[[np.sin(rad), np.cos(rad)]]]) * expand_size
            lines = blk.lines_array() + shifted_vec
            lines[..., 0] = np.clip(lines[..., 0], 0, im_w - 1)
            lines[..., 1] = np.clip(lines[..., 1], 0, im_h - 1)
            blk.lines = lines.astype(np.int64).tolist()
            blk.font_size += expand_size

    return final_blk_list


def visualize_textblocks(canvas, blk_list: List[TextBlock]):
    lw = max(round(sum(canvas.shape) / 2 * 0.003), 2)  # line width
    for ii, blk in enumerate(blk_list):
        bx1, by1, bx2, by2 = blk.xyxy
        cv2.rectangle(canvas, (bx1, by1), (bx2, by2), (127, 255, 127), lw)
        lines = blk.lines_array(dtype=np.int32)
        for jj, line in enumerate(lines):
            cv2.putText(canvas, str(jj), line[0], cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 127, 0), 1)
            cv2.polylines(canvas, [line], True, (0, 127, 255), 2)
        cv2.polylines(canvas, [blk.min_rect()], True, (127, 127, 0), 2)
        center = [int((bx1 + bx2) / 2), int((by1 + by2) / 2)]
        cv2.putText(canvas, str(blk.angle), center, cv2.FONT_HERSHEY_SIMPLEX, 1, (127, 127, 255), 2)
        cv2.putText(
            canvas,
            str(ii),
            (bx1, by1 + lw + 2),
            0,
            lw / 3,
            (255, 127, 127),
            max(lw - 1, 1),
            cv2.LINE_AA,
        )
    return canvas
